"""Mixin for working with SQLite respresentation of a 'Results' object."""

import sqlite3
from enum import Enum
from typing import Literal, Union, Optional


class SQLDataShape(Enum):
    """Enum for the shape of the data in the SQL database."""

    WIDE = "wide"
    LONG = "long"


class ResultsDBMixin:
    """Mixin for interacting with a Results object as if it were a SQL database."""

    def _rows(self):
        """Return the rows of the `Results` object as a list of tuples."""
        for index, result in enumerate(self):
            yield from result.rows(index)

    def export_sql_dump(self, shape: Literal["wide", "long"], filename: str):
        """Export the SQL database to a file.

        :param shape: The shape of the data in the database (wide or long)
        :param filename: The filename to save the database to
        """
        shape_enum = self._get_shape_enum(shape)
        conn = self._db(shape=shape_enum)

        with open(filename, "w") as f:
            for line in conn.iterdump():
                f.write(f"{line}\n")

        conn.close()

    def backup_db_to_file(self, shape: Literal["wide", "long"], filename: str):
        """Backup the in-memory database to a file.


        :param shape: The shape of the data in the database (wide or long)
        :param filename: The filename to save the database to

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.backup_db_to_file(filename="backup.db", shape="long")

        """
        shape_enum = self._get_shape_enum(shape)
        # Source database connection (in-memory)
        source_conn = self._db(shape=shape_enum)

        # Destination database connection (file)
        dest_conn = sqlite3.connect(filename)

        # Backup in-memory database to file
        with source_conn:
            source_conn.backup(dest_conn)

        # Close both connections
        source_conn.close()
        dest_conn.close()

    def _db(self, shape: SQLDataShape, remove_prefix=False):
        """Create a SQLite database in memory and return the connection.

        :param shape: The shape of the data in the database (wide or long)
        :param remove_prefix: Whether to remove the prefix from the column names

        """
        if shape == SQLDataShape.LONG:
            conn = sqlite3.connect(":memory:")

            create_table_query = """
            CREATE TABLE self (
                id INTEGER,
                data_type TEXT,
                key TEXT, 
                value TEXT
            )
            """
            conn.execute(create_table_query)

            list_of_tuples = list(self._rows())
            insert_query = (
                "INSERT INTO self (id, data_type, key, value) VALUES (?, ?, ?, ?)"
            )
            conn.executemany(insert_query, list_of_tuples)
            conn.commit()
            return conn
        elif shape == SQLDataShape.WIDE:
            from sqlalchemy import create_engine

            engine = create_engine("sqlite:///:memory:")
            df = self.to_pandas(remove_prefix=remove_prefix)
            df.to_sql("self", engine, index=False, if_exists="replace")
            return engine.connect()
        else:
            raise Exception("Invalid SQLDataShape")

    def _get_shape_enum(self, shape: Literal["wide", "long"]):
        """Convert the shape string to a SQLDataShape enum."""
        if shape is None:
            raise Exception("Must select either 'wide' or 'long' format")
        elif shape == "wide":
            return SQLDataShape.WIDE
        elif shape == "long":
            return SQLDataShape.LONG
        else:
            raise Exception("Invalid shape: must be either 'long' or 'wide'")

    def sql(
        self,
        query: str,
        shape: Literal["wide", "long"] = "wide",
        remove_prefix: bool = True,
        transpose: bool = None,
        transpose_by: str = None,
        csv: bool = False,
        to_list=False,
        to_latex=False,
        filename: Optional[str] = None,
    ) -> Union["pd.DataFrame", str]:
        """Execute a SQL query and return the results as a DataFrame.

        :param query: The SQL query to execute
        :param shape: The shape of the data in the database (wide or long)
        :param remove_prefix: Whether to remove the prefix from the column names
        :param transpose: Whether to transpose the DataFrame
        :param transpose_by: The column to use as the index when transposing
        :param csv: Whether to return the DataFrame as a CSV string


        Example usage:

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> d = r.sql("select data_type, key, value from self where data_type = 'answer' order by value limit 3", shape="long")
        >>> sorted(list(d['value']))
        ['Good', 'Great', 'Great']

        We can also return the data in wide format.
        Note the use of single quotes to escape the column names, as required by sql.

        >>> from edsl.results import Results
        >>> Results.example().sql("select how_feeling from self", shape = 'wide', remove_prefix=True)
          how_feeling
        0          OK
        1       Great
        2    Terrible
        3          OK
        """
        import pandas as pd

        shape_enum = self._get_shape_enum(shape)

        conn = self._db(shape=shape_enum, remove_prefix=remove_prefix)
        df = pd.read_sql_query(query, conn)

        # Transpose the DataFrame if transpose is True
        if transpose or transpose_by:
            df = pd.DataFrame(df)
            if transpose_by:
                df = df.set_index(transpose_by)
            else:
                df = df.set_index(df.columns[0])
            df = df.transpose()

        if csv and to_list:
            raise Exception("Cannot return both CSV and list")

        if to_list:
            return df.values.tolist()

        if to_latex:
            df.columns = [col.replace("_", " ") for col in df.columns]

            latex_output = df.to_latex(index=False)
            if filename:
                with open(filename, "w") as f:
                    f.write(latex_output)
                return None
            return latex_output

        if csv:
            if filename:
                df.to_csv(filename, index=False)
                return None

            return df.to_csv(index=False)

        return df

    def show_schema(
        self, shape: Literal["wide", "long"], remove_prefix: bool = False
    ) -> None:
        """Show the schema of the Results database.

        :param shape: The shape of the data in the database (wide or long)
        :param remove_prefix: Whether to remove the prefix from the column names

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.show_schema(shape="long")
        Type: table, Name: self, SQL: CREATE TABLE self (
        ...
        <BLANKLINE>
        """
        import pandas as pd

        shape_enum = self._get_shape_enum(shape)
        conn = self._db(shape=shape_enum, remove_prefix=remove_prefix)

        if shape_enum == SQLDataShape.LONG:
            # Query to get the schema of all tables
            query = "SELECT type, name, sql FROM sqlite_master WHERE type='table'"
            cursor = conn.execute(query)
            schema = cursor.fetchall()
            conn.close()

            # Format and return the schema information
            schema_info = ""
            for row in schema:
                schema_info += f"Type: {row[0]}, Name: {row[1]}, SQL: {row[2]}\n"

            print(schema_info)
        elif shape_enum == SQLDataShape.WIDE:
            query = f"PRAGMA table_info(self)"
            schema = pd.read_sql(query, conn)
            # print(schema)
            return schema


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
import warnings


class InferenceServicesCollection:
    added_models = {}

    def __init__(self, services: list[InferenceServiceABC] = None):
        self.services = services or []

    @classmethod
    def add_model(cls, service_name, model_name):
        if service_name not in cls.added_models:
            cls.added_models[service_name] = []
        cls.added_models[service_name].append(model_name)

    @staticmethod
    def _get_service_available(service, warn: bool = False) -> list[str]:
        from_api = True
        try:
            service_models = service.available()
        except Exception as e:
            if warn:
                warnings.warn(
                    f"""Error getting models for {service._inference_service_}. 
                    Check that you have properly stored your Expected Parrot API key and activated remote inference, or stored your own API keys for the language models that you want to use.
                    See https://docs.expectedparrot.com/en/latest/api_keys.html for instructions on storing API keys.
                    Relying on cache.""",
                    UserWarning,
                )
            from edsl.inference_services.models_available_cache import models_available

            service_models = models_available.get(service._inference_service_, [])
            # cache results
            service._models_list_cache = service_models
            from_api = False
        return service_models  # , from_api

    def available(self):
        total_models = []
        for service in self.services:
            service_models = self._get_service_available(service)
            for model in service_models:
                total_models.append([model, service._inference_service_, -1])

            for model in self.added_models.get(service._inference_service_, []):
                total_models.append([model, service._inference_service_, -1])

        sorted_models = sorted(total_models)
        for i, model in enumerate(sorted_models):
            model[2] = i
            model = tuple(model)
        return sorted_models

    def register(self, service):
        self.services.append(service)

    def create_model_factory(self, model_name: str, service_name=None, index=None):
        from edsl.inference_services.TestService import TestService

        if model_name == "test":
            return TestService.create_model(model_name)

        if service_name:
            for service in self.services:
                if service_name == service._inference_service_:
                    return service.create_model(model_name)

        for service in self.services:
            if model_name in self._get_service_available(service):
                if service_name is None or service_name == service._inference_service_:
                    return service.create_model(model_name)

        raise Exception(f"Model {model_name} not found in any of the services")



"""Mixin class for exporting results."""

from functools import wraps
from typing import Literal, Optional, Union

from edsl.results.DatasetExportMixin import DatasetExportMixin


def to_dataset(func):
    """Convert the Results object to a Dataset object before calling the function."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Return the function with the Results object converted to a Dataset object."""
        if self.__class__.__name__ == "Results":
            return func(self.select(), *args, **kwargs)
        else:
            return func(self, *args, **kwargs)

    wrapper._is_wrapped = True
    return wrapper


def decorate_methods_from_mixin(cls, mixin_cls):
    for attr_name, attr_value in mixin_cls.__dict__.items():
        if callable(attr_value) and not attr_name.startswith("__"):
            setattr(cls, attr_name, to_dataset(attr_value))
    return cls


class ResultsExportMixin(DatasetExportMixin):
    """Mixin class for exporting Results objects."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        decorate_methods_from_mixin(cls, DatasetExportMixin)


if __name__ == "__main__":
    # pass
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from typing import List, Optional
import pandas as pd
from edsl.conjure.InputData import InputDataABC
from edsl.conjure.utilities import convert_value


class InputDataCSV(InputDataABC):
    def __init__(self, datafile_name: str, config: Optional[dict] = None, **kwargs):
        if config is None:
            config = {"skiprows": None, "delimiter": ","}

        super().__init__(datafile_name, config, **kwargs)

    def get_df(self) -> pd.DataFrame:
        if not hasattr(self, "_df"):
            self._df = pd.read_csv(
                self.datafile_name,
                skiprows=self.config["skiprows"],
                encoding_errors="ignore",
            )
            float_columns = self._df.select_dtypes(include=["float64"]).columns
            self._df[float_columns] = self._df[float_columns].astype(str)
            self._df.fillna("", inplace=True)
            self._df = self._df.astype(str)
        return self._df

    def get_raw_data(self) -> List[List[str]]:
        data = [
            [convert_value(obs) for obs in v]
            for k, v in self.get_df().to_dict(orient="list").items()
        ]
        return data

    def get_question_texts(self):
        return list(self.get_df().columns)

    def get_question_names(self):
        new_names = [self.naming_function(q) for q in self.question_texts]

        if len(new_names) > len(set(new_names)):
            from collections import Counter

            counter = Counter(new_names)
            for i, name in enumerate(new_names):
                if counter[name] > 1:
                    new_names[i] = name + str(counter[name])
                    counter[name] -= 1
        return new_names



"""This module contains the Question class, which is the base class for all questions in EDSL."""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Type, Optional, List, Callable, Union, TypedDict
import copy

from edsl.exceptions import (
    QuestionResponseValidationError,
    QuestionAnswerValidationError,
    QuestionSerializationError,
)
from edsl.questions.descriptors import QuestionNameDescriptor, QuestionTextDescriptor


from edsl.questions.AnswerValidatorMixin import AnswerValidatorMixin
from edsl.questions.RegisterQuestionsMeta import RegisterQuestionsMeta
from edsl.Base import PersistenceMixin, RichPrintingMixin
from edsl.BaseDiff import BaseDiff, BaseDiffCollection

from edsl.questions.SimpleAskMixin import SimpleAskMixin
from edsl.questions.QuestionBasePromptsMixin import QuestionBasePromptsMixin
from edsl.questions.QuestionBaseGenMixin import QuestionBaseGenMixin
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version


class QuestionBase(
    PersistenceMixin,
    RichPrintingMixin,
    SimpleAskMixin,
    QuestionBasePromptsMixin,
    QuestionBaseGenMixin,
    ABC,
    AnswerValidatorMixin,
    metaclass=RegisterQuestionsMeta,
):
    """ABC for the Question class. All questions inherit from this class.
    Some of the constraints on child questions are defined in the RegisterQuestionsMeta metaclass.
    """

    question_name: str = QuestionNameDescriptor()
    question_text: str = QuestionTextDescriptor()

    _answering_instructions = None
    _question_presentation = None

    @property
    def response_model(self) -> type["BaseModel"]:
        if self._response_model is not None:
            return self._response_model
        else:
            return self.create_response_model()

    # region: Validation and simulation methods
    @property
    def response_validator(self) -> "ResponseValidatorBase":
        """Return the response validator."""
        params = (
            {
                "response_model": self.response_model,
            }
            | {k: getattr(self, k) for k in self.validator_parameters}
            | {"exception_to_throw": getattr(self, "exception_to_throw", None)}
            | {"override_answer": getattr(self, "override_answer", None)}
        )
        return self.response_validator_class(**params)

    @property
    def validator_parameters(self) -> list[str]:
        """Return the parameters required for the response validator.

        >>> from edsl import QuestionMultipleChoice as Q
        >>> Q.example().validator_parameters
        ['question_options', 'use_code']

        """
        return self.response_validator_class.required_params

    @property
    def fake_data_factory(self):
        """Return the fake data factory."""
        if not hasattr(self, "_fake_data_factory"):
            from polyfactory.factories.pydantic_factory import ModelFactory

            class FakeData(ModelFactory[self.response_model]):
                ...

            self._fake_data_factory = FakeData
        return self._fake_data_factory

    def _simulate_answer(self, human_readable: bool = False) -> dict:
        """Simulate a valid answer for debugging purposes (what the validator expects).
        >>> from edsl import QuestionFreeText as Q
        >>> Q.example()._simulate_answer()
        {'answer': '...', 'generated_tokens': ...}
        """
        simulated_answer = self.fake_data_factory.build().dict()
        if human_readable and hasattr(self, "question_options") and self.use_code:
            simulated_answer["answer"] = [
                self.question_options[index] for index in simulated_answer["answer"]
            ]
        return simulated_answer

    class ValidatedAnswer(TypedDict):
        answer: Any
        comment: Optional[str]
        generated_tokens: Optional[str]

    def _validate_answer(
        self, answer: dict, replacement_dict: dict = None
    ) -> ValidatedAnswer:
        """Validate the answer.
        >>> from edsl.exceptions import QuestionAnswerValidationError
        >>> from edsl import QuestionFreeText as Q
        >>> Q.example()._validate_answer({'answer': 'Hello', 'generated_tokens': 'Hello'})
        {'answer': 'Hello', 'generated_tokens': 'Hello'}
        """

        return self.response_validator.validate(answer, replacement_dict)

    # endregion

    # region: Serialization methods
    @property
    def name(self) -> str:
        "Helper function so questions and instructions can use the same access method"
        return self.question_name

    def __hash__(self) -> int:
        """Return a hash of the question.

        >>> from edsl import QuestionFreeText as Q
        >>> hash(Q.example())
        1144312636257752766
        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self._to_dict())

    @property
    def data(self) -> dict:
        """Return a dictionary of question attributes **except** for question_type.

        >>> from edsl import QuestionFreeText as Q
        >>> Q.example().data
        {'question_name': 'how_are_you', 'question_text': 'How are you?'}
        """
        exclude_list = [
            "question_type",
            "_include_comment",
            "_fake_data_factory",
            "_use_code",
            "_answering_instructions",
            "_question_presentation",
            "_model_instructions",
        ]
        candidate_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_") and k not in exclude_list
        }

        if "func" in candidate_data:
            func = candidate_data.pop("func")
            import inspect

            candidate_data["function_source_code"] = inspect.getsource(func)

        return candidate_data

    def _to_dict(self):
        """Convert the question to a dictionary that includes the question type (used in deserialization).

        >>> from edsl import QuestionFreeText as Q; Q.example()._to_dict()
        {'question_name': 'how_are_you', 'question_text': 'How are you?', 'question_type': 'free_text'}
        """
        candidate_data = self.data.copy()
        candidate_data["question_type"] = self.question_type
        return candidate_data

    @add_edsl_version
    def to_dict(self) -> dict[str, Any]:
        """Convert the question to a dictionary that includes the question type (used in deserialization).
        >>> from edsl import QuestionFreeText as Q; Q.example().to_dict()
        {'question_name': 'how_are_you', 'question_text': 'How are you?', 'question_type': 'free_text', 'edsl_version': '...'}
        """
        return self._to_dict()

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> Type[QuestionBase]:
        """Construct a question object from a dictionary created by that question's `to_dict` method."""
        local_data = data.copy()

        try:
            question_type = local_data.pop("question_type")
            if question_type == "linear_scale":
                # This is a fix for issue https://github.com/expectedparrot/edsl/issues/165
                options_labels = local_data.get("option_labels", None)
                if options_labels:
                    options_labels = {
                        int(key): value for key, value in options_labels.items()
                    }
                    local_data["option_labels"] = options_labels
        except:
            raise QuestionSerializationError(
                f"Data does not have a 'question_type' field (got {data})."
            )
        from edsl.questions.question_registry import get_question_class

        try:
            question_class = get_question_class(question_type)
        except ValueError:
            raise QuestionSerializationError(
                f"No question registered with question_type {question_type}"
            )

        if "model_instructions" in local_data:
            model_instructions = local_data.pop("model_instructions")
            new_q = question_class(**local_data)
            new_q.model_instructions = model_instructions
            return new_q

        return question_class(**local_data)

    # endregion

    # region: Running methods
    @classmethod
    def _get_test_model(self, canned_response: Optional[str] = None) -> "LanguageModel":
        """Get a test model for the question."""
        from edsl.language_models import LanguageModel

        return LanguageModel.example(canned_response=canned_response, test_model=True)

    @classmethod
    def run_example(
        cls,
        show_answer: bool = True,
        model: Optional["LanguageModel"] = None,
        cache=False,
        **kwargs,
    ):
        """Run an example of the question.
        >>> from edsl.language_models import LanguageModel
        >>> from edsl import QuestionFreeText as Q
        >>> m = Q._get_test_model(canned_response = "Yo, what's up?")
        >>> m.execute_model_call("", "")
        {'message': [{'text': "Yo, what's up?"}], 'usage': {'prompt_tokens': 1, 'completion_tokens': 1}}
        >>> Q.run_example(show_answer = True, model = m)
        ┏━━━━━━━━━━━━━━━━┓
        ┃ answer         ┃
        ┃ .how_are_you   ┃
        ┡━━━━━━━━━━━━━━━━┩
        │ Yo, what's up? │
        └────────────────┘
        """
        if model is None:
            from edsl import Model

            model = Model()
        results = cls.example(**kwargs).by(model).run(cache=cache)
        if show_answer:
            results.select("answer.*").print()
        else:
            return results

    def __call__(self, just_answer=True, model=None, agent=None, **kwargs):
        """Call the question.


        >>> from edsl import QuestionFreeText as Q
        >>> m = Q._get_test_model(canned_response = "Yo, what's up?")
        >>> q = Q(question_name = "color", question_text = "What is your favorite color?")
        >>> q(model = m)
        "Yo, what's up?"

        """
        survey = self.to_survey()
        results = survey(model=model, agent=agent, **kwargs, cache=False)
        if just_answer:
            return results.select(f"answer.{self.question_name}").first()
        else:
            return results

    def run(self, *args, **kwargs) -> "Results":
        """Turn a single question into a survey and runs it."""
        from edsl.surveys.Survey import Survey

        s = self.to_survey()
        return s.run(*args, **kwargs)

    async def run_async(
        self,
        just_answer: bool = True,
        model: Optional["Model"] = None,
        agent: Optional["Agent"] = None,
        **kwargs,
    ) -> Union[Any, "Results"]:
        """Call the question asynchronously.

        >>> import asyncio
        >>> from edsl import QuestionFreeText as Q
        >>> m = Q._get_test_model(canned_response = "Blue")
        >>> q = Q(question_name = "color", question_text = "What is your favorite color?")
        >>> async def test_run_async(): result = await q.run_async(model=m); print(result)
        >>> asyncio.run(test_run_async())
        Blue
        """
        survey = self.to_survey()
        results = await survey.run_async(model=model, agent=agent, **kwargs)
        if just_answer:
            return results.select(f"answer.{self.question_name}").first()
        else:
            return results

    # endregion

    # region: Magic methods
    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html

        data = self.to_dict()
        try:
            _ = data.pop("edsl_version")
            _ = data.pop("edsl_class_name")
        except KeyError:
            print("Serialized question lacks edsl version, but is should have it.")

        return data_to_html(data)

    def __getitem__(self, key: str) -> Any:
        """Get an attribute of the question so it can be treated like a dictionary.

        >>> from edsl.questions import QuestionFreeText as Q
        >>> Q.example()['question_text']
        'How are you?'
        """
        return getattr(self, key)

    def __repr__(self) -> str:
        """Return a string representation of the question. Should be able to be used to reconstruct the question.

        >>> from edsl import QuestionFreeText as Q
        >>> repr(Q.example())
        'Question(\\'free_text\\', question_name = \"""how_are_you\""", question_text = \"""How are you?\""")'
        """
        items = [
            f'{k} = """{v}"""' if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        question_type = self.to_dict().get("question_type", "None")
        return f"Question('{question_type}', {', '.join(items)})"

    def __eq__(self, other: Union[Any, Type[QuestionBase]]) -> bool:
        """Check if two questions are equal. Equality is defined as having the .to_dict().

        >>> from edsl import QuestionFreeText as Q
        >>> q1 = Q.example()
        >>> q2 = Q.example()
        >>> q1 == q2
        True
        >>> q1.question_text = "How are you John?"
        >>> q1 == q2
        False

        """
        if not isinstance(other, QuestionBase):
            return False
        return self.to_dict() == other.to_dict()

    def __sub__(self, other) -> BaseDiff:
        """Return the difference between two objects.
        >>> from edsl import QuestionFreeText as Q
        >>> q1 = Q.example()
        >>> q2 = q1.copy()
        >>> q2.question_text = "How are you John?"
        >>> diff = q1 - q2
        """

        return BaseDiff(other, self)

    # TODO: Throws an error that should be addressed at QuestionFunctional
    def __add__(self, other_question_or_diff):
        """
        Compose two questions into a single question.

        TODO: Probably getting deprecated.

        """
        if isinstance(other_question_or_diff, BaseDiff) or isinstance(
            other_question_or_diff, BaseDiffCollection
        ):
            return other_question_or_diff.apply(self)

        # from edsl.questions import compose_questions
        # return compose_questions(self, other_question_or_diff)

    # def _validate_response(self, response):
    #     """Validate the response from the LLM. Behavior depends on the question type."""
    #     if "answer" not in response:
    #         raise QuestionResponseValidationError(
    #             "Response from LLM does not have an answer"
    #         )
    #     return response

    def _translate_answer_code_to_answer(
        self, answer, scenario: Optional["Scenario"] = None
    ):
        """There is over-ridden by child classes that ask for codes."""
        return answer

    # endregion

    # region: Forward methods
    def add_question(self, other: QuestionBase) -> "Survey":
        """Add a question to this question by turning them into a survey with two questions.

        >>> from edsl.questions import QuestionFreeText as Q
        >>> from edsl.questions import QuestionMultipleChoice as QMC
        >>> s = Q.example().add_question(QMC.example())
        >>> len(s.questions)
        2
        """
        from edsl.surveys.Survey import Survey

        s = Survey([self, other])
        return s

    def to_survey(self) -> "Survey":
        """Turn a single question into a survey.
        >>> from edsl import QuestionFreeText as Q
        >>> Q.example().to_survey().questions[0].question_name
        'how_are_you'
        """
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s

    def by(self, *args) -> "Jobs":
        """Turn a single question into a survey and then a Job."""
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s.by(*args)

    # endregion

    # region: Display methods
    def print(self):
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()))

    def human_readable(self) -> str:
        """Print the question in a human readable format.

        >>> from edsl.questions import QuestionFreeText
        >>> QuestionFreeText.example().human_readable()
        'Question Type: free_text\\nQuestion: How are you?'
        """
        lines = []
        lines.append(f"Question Type: {self.question_type}")
        lines.append(f"Question: {self.question_text}")
        if hasattr(self, "question_options"):
            lines.append("Please name the option you choose from the following.:")
            for index, option in enumerate(self.question_options):
                lines.append(f"{option}")
        return "\n".join(lines)

    def html(
        self,
        scenario: Optional[dict] = None,
        agent: Optional[dict] = {},
        answers: Optional[dict] = None,
        include_question_name: bool = False,
        height: Optional[int] = None,
        width: Optional[int] = None,
        iframe=False,
    ):
        """Return the question in HTML format."""
        from jinja2 import Template

        if scenario is None:
            scenario = {}

        prior_answers_dict = {}

        if isinstance(answers, dict):
            for key, value in answers.items():
                if not key.endswith("_comment") and not key.endswith(
                    "_generated_tokens"
                ):
                    prior_answers_dict[key] = {"answer": value}

        # breakpoint()

        base_template = """
        <div id="{{ question_name }}" class="survey_question" data-type="{{ question_type }}">
            {% if include_question_name %}
            <p>question_name: {{ question_name }}</p>
            {% endif %}
            <p class="question_text">{{ question_text }}</p>
            {{ question_content }}
        </div>
        """
        if not hasattr(self, "question_type"):
            self.question_type = "unknown"

        if hasattr(self, "question_html_content"):
            question_content = self.question_html_content
        else:
            question_content = Template("")

        base_template = Template(base_template)

        context = {
            "scenario": scenario,
            "agent": agent,
        } | prior_answers_dict

        # Render the question text
        try:
            question_text = Template(self.question_text).render(context)
        except Exception as e:
            print(
                f"Error rendering question: question_text = {self.question_text}, error = {e}"
            )
            question_text = self.question_text

        try:
            question_content = Template(question_content).render(context)
        except Exception as e:
            print(
                f"Error rendering question: question_content = {question_content}, error = {e}"
            )
            question_content = question_content

        try:
            params = {
                "question_name": self.question_name,
                "question_text": question_text,
                "question_type": self.question_type,
                "question_content": question_content,
                "include_question_name": include_question_name,
            }
        except Exception as e:
            raise ValueError(
                f"Error rendering question: params = {params}, error = {e}"
            )
        rendered_html = base_template.render(**params)

        if iframe:
            import html
            from IPython.display import display, HTML

            height = height or 200
            width = width or 600
            escaped_output = html.escape(rendered_html)
            # escaped_output = rendered_html
            iframe = f""""
            <iframe srcdoc="{ escaped_output }" style="width: {width}px; height: {height}px;"></iframe>
            """
            display(HTML(iframe))
            return None

        return rendered_html

    @classmethod
    def example_model(cls):
        from edsl import Model

        q = cls.example()
        m = Model("test", canned_response=cls._simulate_answer(q)["answer"])

        return m

    @classmethod
    def example_results(cls):
        m = cls.example_model()
        q = cls.example()
        return q.by(m).run(cache=False)

    def rich_print(self):
        """Print the question in a rich format."""
        from rich.table import Table

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Question Name", style="dim")
        table.add_column("Question Type")
        table.add_column("Question Text")
        table.add_column("Options")

        question = self
        if hasattr(question, "question_options"):
            options = ", ".join([str(o) for o in question.question_options])
        else:
            options = "None"
        table.add_row(
            question.question_name,
            question.question_type,
            question.question_text,
            options,
        )
        return table

    # endregion


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from textwrap import dedent
import random
from typing import List, TypeVar, Generator, Optional
from edsl.auto.StageBase import StageBase
from edsl.conjure.naming_utilities import sanitize_string
from edsl import Agent, Survey, Model, Cache, AgentList
from edsl import QuestionFreeText, Scenario
from edsl import QuestionMultipleChoice, Scenario, Agent, ScenarioList

StageClassType = TypeVar("StageClassType", bound=StageBase)


def gen_pipeline(stages_list: List[StageClassType]) -> StageBase:
    """Takes as input a list of Stage classes & returns a pipeline of instantiated stages.
    A pipeline is a linked list of stages where each stage has a next_stage attribute.

    """
    pipeline = stages_list[0]()
    last_stage = pipeline
    for stage in stages_list[1:]:
        while last_stage.next_stage is not None:  # find the end of the pipeline
            last_stage = last_stage.next_stage
        stage_to_add = stage()
        last_stage.next_stage = stage_to_add
    return pipeline


q_eligibility = QuestionMultipleChoice(
    question_text=dedent(
        """\
        Consider this set of question: '{{ questions }}'.
        Consider this persona: '{{ persona }}'.
        Would this persona be able to answer all of these questions?
        """
    ),
    question_options=["No", "Yes"],
    question_name="eligibility",
)


def agent_list_eligibility(
    agent_list: AgentList,
    survey: Optional[Survey] = None,
    model: Optional[Model] = None,
    cache: Optional[Cache] = None,
) -> List[bool]:
    """
    Returns whether each agent in a list is elgible for a survey i.e., can answer every question.

    >>> from edsl.language_models import LanguageModel
    >>> m = LanguageModel.example(canned_response = "1", test_model = True)
    >>> agent_list_eligibility(AgentList.example())
    [True, True]
    >>> agent_list_eligibility(AgentList.example().add_trait('persona', 2*["Cool dude"]), survey = Survey.example(), model = m)
    [True, True]
    """
    if survey is None:
        return [True] * len(agent_list)
    if "persona" not in agent_list.all_traits:
        raise ValueError(
            f"Each agent needs to have a persona attribute; traits are {agent_list.all_traits}"
        )
    sl = agent_list.select("persona").to_scenario_list()
    sl.add_value("questions", [q.question_text for q in survey._questions])
    results = q_eligibility.by(sl).by(model).run(cache=cache)
    return [r == "Yes" for r in results.select("eligibility").to_list()]


def agent_eligibility(
    agent: Agent,
    survey: Survey,
    model: Optional[Model] = None,
    cache: Optional[Cache] = None,
) -> bool:
    """NB: This could be parallelized.

    >>> from edsl.language_models import LanguageModel
    >>> m = LanguageModel.example(canned_response = "1", test_model = True)
    >>> agent_eligibility(agent = Agent.example().add_trait({'persona': "Persona"}), survey = Survey.example(), model = m)
    True

    """
    model = model or Model()

    questions = [q.question_text for q in survey._questions]
    persona = agent.traits["persona"]
    return (
        q_eligibility(model=model, questions=questions, persona=persona, cache=cache)
        == "Yes"
    )
    # results = (
    #     q.by(model)
    #     .by(Scenario({"questions": questions, "persona": persona}))
    #     .run(cache=cache)
    # )
    # return results.select("eligibility").first() == "Yes"


def gen_agent_traits(dimension_dict: dict, seed_value: Optional[str] = None):
    """
    >>> dimension_dict = {'attitude':['positive', 'negative']}
    >>> ag = gen_agent_traits(dimension_dict)
    >>> a = next(ag)
    >>> a == {'attitude': 'positive'} or a == {'attitude': 'negative'}
    True
    >>> len([next(ag) for _ in range(100)])
    100
    """
    if seed_value is None:
        seed_value = "edsl"

    random.seed(seed_value)

    while True:
        new_agent_traits = {}
        for key, list_of_values in dimension_dict.items():
            new_agent_traits[key] = random.choice(list_of_values)
        yield new_agent_traits


def agent_generator(
    persona: str,
    dimension_dict: dict,
    model: Optional[Model] = None,
    cache: Optional["Cache"] = None,
) -> Generator["Results", None, None]:
    """
    >>> from edsl.language_models import LanguageModel
    >>> m = LanguageModel.example(canned_response = "This is a cool dude.", test_model = True)
    >>> ag = agent_generator(persona = "Base person", dimension_dict = {'attitude':['Positive', 'Negative']}, model = m)
    >>> next(ag).select('new_agent_persona').first()
    'This is a cool dude.'
    >>> next(ag).select('new_agent_persona').first()
    'This is a cool dude.'
    """

    if model is None:
        model = Model()

    q = QuestionFreeText(
        question_text=dedent(
            """\
    Consider this persona: '{{ persona }}'.
    Now imagine writing a new persona with these traits: 
    '{{ new_agent_traits }}'
    Please write this persona as a narrative.
    """
        ),
        question_name="new_agent_persona",
    )
    agent_trait_generator = gen_agent_traits(dimension_dict)
    codebook = {sanitize_string(k): k for k in dimension_dict.keys()}
    while True:
        new_agent_traits = next(agent_trait_generator)
        yield q(
            persona=persona,
            new_agent_traits=new_agent_traits,
            codebook=codebook,
            just_answer=False,
            cache=cache,
            model=model,
        )


def create_agents(
    agent_generator: Generator["Results", None, None],
    survey: Optional[Survey] = None,
    num_agents=11,
) -> AgentList:
    """
    >>> from edsl.language_models import LanguageModel
    >>> m = LanguageModel.example(canned_response = "This is a cool dude.", test_model = True)
    >>> ag = agent_generator(persona = "Base person", dimension_dict = {'attitude':['Positive', 'Negative']}, model = m)
    >>> new_agent_list = create_agents(agent_generator = ag)
    >>> new_agent_list

    """
    agent_list = AgentList([])

    MAX_ITERATIONS_MULTIPLIER = 2
    iterations = 0

    while len(agent_list) < num_agents:
        iterations += 1
        candidate_agent = next(agent_generator)
        codebook = candidate_agent.select("codebook").to_list()[0]

        koobedoc = {v: k for k, v in codebook.items()}
        persona = candidate_agent.select("new_agent_persona").to_list()[0]
        traits = candidate_agent.select("new_agent_traits").to_list()[0]
        new_traits = {koobedoc[key]: value for key, value in traits.items()} | {
            "persona": persona
        }
        agent = Agent(traits=new_traits, codebook=codebook)
        if survey is not None:
            if agent_eligibility(agent, survey):
                agent_list.append(agent)
            else:
                print("Agent not eligible")
        else:
            agent_list.append(agent)

        if iterations > MAX_ITERATIONS_MULTIPLIER * num_agents:
            raise Exception("Too many failures")

    return agent_list


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # from edsl.language_models import LanguageModel

    # m = LanguageModel.example(canned_response="This is a cool dude.", test_model=True)
    # ag = agent_generator(
    #     persona="Base person",
    #     dimension_dict={"attitude": ["Positive", "Negative"]},
    #     model=m,
    # )
    # example = [next(ag).select("new_agent_persona").first() for _ in range(10)]
    # dimension_dict = {"attitude": ["positive", "negative"]}
    # ag = gen_agent_traits(dimension_dict)
    # example = [next(ag) for _ in range(100)]



from __future__ import annotations

import time
from dataclasses import dataclass, asdict

from typing import List, DefaultDict, Optional, Type, Literal
from collections import UserDict, defaultdict

from rich.text import Text
from rich.box import SIMPLE
from rich.table import Table
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn
from rich.layout import Layout
from rich.console import Group
from rich import box

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.tokens.TokenUsage import TokenUsage
from edsl.enums import get_token_pricing
from edsl.jobs.tasks.task_status_enum import TaskStatus

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]

from edsl.jobs.interviews.InterviewStatistic import InterviewStatistic
from edsl.jobs.interviews.InterviewStatisticsCollection import (
    InterviewStatisticsCollection,
)
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage


@dataclass
class ModelInfo:
    model_name: str
    TPM_limit_k: float
    RPM_limit_k: float
    num_tasks_waiting: int
    token_usage_info: dict


@dataclass
class ModelTokenUsageStats:
    token_usage_type: str
    details: List[dict]
    cost: str


class Stats:
    def elapsed_time(self):
        InterviewStatistic("elapsed_time", value=elapsed_time, digits=1, units="sec.")


class JobsRunnerStatus:
    def __init__(
        self, jobs_runner: "JobsRunnerAsyncio", n: int, refresh_rate: float = 0.25
    ):
        self.jobs_runner = jobs_runner
        self.start_time = time.time()
        self.completed_interviews = []
        self.refresh_rate = refresh_rate
        self.statistics = [
            "elapsed_time",
            "total_interviews_requested",
            "completed_interviews",
            #            "percent_complete",
            "average_time_per_interview",
            #            "task_remaining",
            "estimated_time_remaining",
            "exceptions",
            "unfixed_exceptions",
            "throughput",
        ]
        self.num_total_interviews = n * len(self.jobs_runner.interviews)

        self.distinct_models = list(
            set(i.model.model for i in self.jobs_runner.interviews)
        )

        self.completed_interview_by_model = defaultdict(list)

    def add_completed_interview(self, result):
        self.completed_interviews.append(result.interview_hash)

        relevant_model = result.model.model
        self.completed_interview_by_model[relevant_model].append(result.interview_hash)

    def _compute_statistic(self, stat_name: str):
        completed_tasks = self.completed_interviews
        elapsed_time = time.time() - self.start_time
        interviews = self.jobs_runner.total_interviews

        stat_definitions = {
            "elapsed_time": lambda: InterviewStatistic(
                "elapsed_time", value=elapsed_time, digits=1, units="sec."
            ),
            "total_interviews_requested": lambda: InterviewStatistic(
                "total_interviews_requested", value=len(interviews), units=""
            ),
            "completed_interviews": lambda: InterviewStatistic(
                "completed_interviews", value=len(completed_tasks), units=""
            ),
            "percent_complete": lambda: InterviewStatistic(
                "percent_complete",
                value=(
                    len(completed_tasks) / len(interviews) * 100
                    if len(interviews) > 0
                    else 0
                ),
                digits=1,
                units="%",
            ),
            "average_time_per_interview": lambda: InterviewStatistic(
                "average_time_per_interview",
                value=elapsed_time / len(completed_tasks) if completed_tasks else 0,
                digits=2,
                units="sec.",
            ),
            "task_remaining": lambda: InterviewStatistic(
                "task_remaining", value=len(interviews) - len(completed_tasks), units=""
            ),
            "estimated_time_remaining": lambda: InterviewStatistic(
                "estimated_time_remaining",
                value=(
                    (len(interviews) - len(completed_tasks))
                    * (elapsed_time / len(completed_tasks))
                    if len(completed_tasks) > 0
                    else 0
                ),
                digits=1,
                units="sec.",
            ),
            "exceptions": lambda: InterviewStatistic(
                "exceptions",
                value=sum(len(i.exceptions) for i in interviews),
                units="",
            ),
            "unfixed_exceptions": lambda: InterviewStatistic(
                "unfixed_exceptions",
                value=sum(i.exceptions.num_unfixed() for i in interviews),
                units="",
            ),
            "throughput": lambda: InterviewStatistic(
                "throughput",
                value=len(completed_tasks) / elapsed_time if elapsed_time > 0 else 0,
                digits=2,
                units="interviews/sec.",
            ),
        }
        return stat_definitions[stat_name]()

    def create_progress_bar(self):
        return Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("{task.completed}/{task.total}"),
        )

    def generate_model_queues_table(self):
        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Info", style="cyan")
        table.add_column("Value", style="magenta")
        # table.add_row("Bucket collection", str(self.jobs_runner.bucket_collection))
        for model, bucket in self.jobs_runner.bucket_collection.items():
            table.add_row(Text(model.model, style="bold blue"), "")
            bucket_types = ["requests_bucket", "tokens_bucket"]
            for bucket_type in bucket_types:
                table.add_row(Text(" " + bucket_type, style="green"), "")
                # table.add_row(
                #     f"  Current level (capacity = {round(getattr(bucket, bucket_type).capacity, 3)})",
                #     str(round(getattr(bucket, bucket_type).tokens, 3)),
                # )
                num_requests = getattr(bucket, bucket_type).num_requests
                num_released = getattr(bucket, bucket_type).num_released
                tokens_returned = getattr(bucket, bucket_type).tokens_returned
                # table.add_row(
                #     f"  Requested",
                #     str(num_requests),
                # )
                # table.add_row(
                #     f"  Completed",
                #     str(num_released),
                # )
                table.add_row(
                    "  Completed vs. Requested", f"{num_released} vs. {num_requests}"
                )
                table.add_row(
                    "  Added tokens (from cache)",
                    str(tokens_returned),
                )
                if bucket_type == "tokens_bucket":
                    rate_name = "TPM"
                else:
                    rate_name = "RPM"
                target_rate = round(getattr(bucket, bucket_type).target_rate, 1)
                table.add_row(
                    f"  Empirical {rate_name} (target = {target_rate})",
                    str(round(getattr(bucket, bucket_type).get_throughput(), 0)),
                )

        return table

    def generate_layout(self):
        progress = self.create_progress_bar()
        task_ids = []
        for model in self.distinct_models:
            task_id = progress.add_task(
                f"[cyan]{model}...",
                total=int(self.num_total_interviews / len(self.distinct_models)),
            )
            task_ids.append((model, task_id))

        progress_height = min(5, 2 + len(self.distinct_models))
        layout = Layout()

        # Create the top row with only the progress panel
        layout.split_column(
            Layout(
                Panel(
                    progress,
                    title="Interview Progress",
                    border_style="cyan",
                    box=box.ROUNDED,
                ),
                name="progress",
                size=progress_height,  # Adjusted size
            ),
            Layout(name="bottom_row"),  # Adjusted size
        )

        # Split the bottom row into two columns for metrics and model queues
        layout["bottom_row"].split_row(
            Layout(
                Panel(
                    self.generate_metrics_table(),
                    title="Metrics",
                    border_style="magenta",
                    box=box.ROUNDED,
                ),
                name="metrics",
            ),
            Layout(
                Panel(
                    self.generate_model_queues_table(),
                    title="Model Queues",
                    border_style="yellow",
                    box=box.ROUNDED,
                ),
                name="model_queues",
            ),
        )

        return layout, progress, task_ids

    def generate_metrics_table(self):
        table = Table(show_header=True, header_style="bold magenta", box=box.SIMPLE)
        table.add_column("Metric", style="cyan", no_wrap=True)
        table.add_column("Value", justify="right")

        for stat_name in self.statistics:
            pretty_name, value = list(self._compute_statistic(stat_name).items())[0]
            # breakpoint()
            table.add_row(pretty_name, value)
        return table

    def update_progress(self, stop_event):
        layout, progress, task_ids = self.generate_layout()

        with Live(
            layout, refresh_per_second=int(1 / self.refresh_rate), transient=True
        ) as live:
            while (
                len(self.completed_interviews) < len(self.jobs_runner.total_interviews)
                and not stop_event.is_set()
            ):
                completed_tasks = len(self.completed_interviews)
                total_tasks = len(self.jobs_runner.total_interviews)

                for model, task_id in task_ids:
                    completed_tasks = len(self.completed_interview_by_model[model])
                    progress.update(
                        task_id,
                        completed=completed_tasks,
                        description=f"[cyan]Conducting interviews for {model}...",
                    )

                layout["metrics"].update(
                    Panel(
                        self.generate_metrics_table(),
                        title="Metrics",
                        border_style="magenta",
                        box=box.ROUNDED,
                    )
                )
                layout["model_queues"].update(
                    Panel(
                        self.generate_model_queues_table(),
                        title="Final Model Queues",
                        border_style="yellow",
                        box=box.ROUNDED,
                    )
                )

                time.sleep(self.refresh_rate)

            # Final update
            for model, task_id in task_ids:
                completed_tasks = len(self.completed_interview_by_model[model])
                progress.update(
                    task_id,
                    completed=completed_tasks,
                    description=f"[cyan]Conducting interviews for {model}...",
                )

            layout["metrics"].update(
                Panel(
                    self.generate_metrics_table(),
                    title="Final Metrics",
                    border_style="magenta",
                    box=box.ROUNDED,
                )
            )
            live.update(layout)
            time.sleep(1)  # Show final state for 1 second


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from openai import AsyncOpenAI
import asyncio

client = AsyncOpenAI(base_url="http://127.0.0.1:8000/v1", api_key="fake_key")


async def main():
    response = await client.chat.completions.create(
        model="gpt-3.5-turbo", messages=[{"role": "user", "content": "Question XX42"}]
    )
    print(response)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Union, Optional, List, Generator, Dict
from edsl.questions import QuestionBase

from edsl.utilities.decorators import add_edsl_version, remove_edsl_version


class Instruction:
    def __init__(
        self, name, text, preamble="You were given the following instructions:"
    ):
        self.name = name
        self.text = text
        self.preamble = preamble

    def __str__(self):
        return self.text

    def __repr__(self):
        return """Instruction(name="{}", text="{}")""".format(self.name, self.text)

    def _to_dict(self):
        return {
            "name": self.name,
            "text": self.text,
            "edsl_class_name": "Instruction",
            "preamble": self.preamble,
        }

    def add_question(self, question) -> "Survey":
        from edsl import Survey

        return Survey([self, question])

    @add_edsl_version
    def to_dict(self):
        return self._to_dict()

    def __hash__(self) -> int:
        """Return a hash of the question."""
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self._to_dict())

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data):
        return cls(
            data["name"],
            data["text"],
            data.get("preamble", "You were given the following instructions:"),
        )



from typing import Union, List


class QuestionOptionMixin:
    @property
    def question_options(self):
        if not hasattr(self, "_question_options"):
            self.question_options = None
        return self._question_options

    @question_options.setter
    def question_options(self, value):
        if value is None:
            value = [self._get_question_options(qn) for qn in self.question_names]
        self._question_options = value

    def _get_question_options(self, question_name) -> Union[List[str], None]:
        """Return the options for a question.

        >>> from edsl.conjure.InputData import InputDataABC
        >>> id = InputDataABC.example()
        >>> sorted(id._get_question_options('morning'))
        ['1', '4']

        """
        qt = self.question_statistics(question_name)
        idx = self.question_names.index(question_name)
        question_type = self.question_types[idx]
        if question_type == "multiple_choice":
            return [str(o) for o in qt.unique_responses]
        else:
            if question_type == "multiple_choice_with_other":
                options = self.unique_responses_more_than_k(2)[
                    self.question_names.index(question_name)
                ] + [self.OTHER_STRING]
                return [str(o) for o in options]
            else:
                return None

    def order_options(self) -> None:
        """Order the options for multiple choice questions using an LLM."""
        from edsl import QuestionList, ScenarioList
        import textwrap

        scenarios = (
            ScenarioList.from_list("example_question_name", self.question_names)
            .add_list("example_question_text", self.question_texts)
            .add_list("example_question_type", self.question_types)
            .add_list("example_question_options", self.question_options)
        ).filter(
            'example_question_type == "multiple_choice" or example_question_type == "multiple_choice_with_other"'
        )

        question = QuestionList(
            question_text=textwrap.dedent(
                """\
            We have a survey question: `{{ example_question_text }}`.
            
            The survey had following options: '{{ example_question_options }}'.
            The options might be out of order. Please put them in the correct order.
            If there is not natural order, just put then in order they were presented.
            """
            ),
            question_name="ordering",
        )
        proposed_ordering = question.by(scenarios).run()
        d = dict(
            proposed_ordering.select("example_question_name", "ordering").to_list()
        )
        self._question_options = [d.get(qn, None) for qn in self.question_names]


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import aiohttp
import json
import os
import requests
from typing import Any, Optional, Union, Literal
from uuid import UUID
import edsl
from edsl import CONFIG, CacheEntry, Jobs, Survey
from edsl.exceptions.coop import CoopNoUUIDError, CoopServerResponseError
from edsl.coop.utils import (
    EDSLObject,
    ObjectRegistry,
    ObjectType,
    RemoteJobStatus,
    VisibilityType,
)


class Coop:
    """
    Client for the Expected Parrot API.
    """

    def __init__(self, api_key: str = None, url: str = None) -> None:
        """
        Initialize the client.
        - Provide an API key directly, or through an env variable.
        - Provide a URL directly, or use the default one.
        """
        self.api_key = api_key or os.getenv("EXPECTED_PARROT_API_KEY")
        self.url = url or CONFIG.EXPECTED_PARROT_URL
        if self.url.endswith("/"):
            self.url = self.url[:-1]
        self._edsl_version = edsl.__version__

    ################
    # BASIC METHODS
    ################
    @property
    def headers(self) -> dict:
        """
        Return the headers for the request.
        """
        headers = {}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        else:
            headers["Authorization"] = f"Bearer None"
        return headers

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
        timeout: Optional[float] = 5,
    ) -> requests.Response:
        """
        Send a request to the server and return the response.
        """
        url = f"{self.url}/{uri}"
        method = method.upper()
        if payload is None:
            timeout = 20
        elif (
            method.upper() == "POST"
            and "json_string" in payload
            and payload.get("json_string") is not None
        ):
            timeout = max(20, (len(payload.get("json_string", "")) // (1024 * 1024)))
        try:
            if method in ["GET", "DELETE"]:
                response = requests.request(
                    method, url, params=params, headers=self.headers, timeout=timeout
                )
            elif method in ["POST", "PATCH"]:
                response = requests.request(
                    method,
                    url,
                    params=params,
                    json=payload,
                    headers=self.headers,
                    timeout=timeout,
                )
            else:
                raise Exception(f"Invalid {method=}.")
        except requests.ConnectionError:
            raise requests.ConnectionError(f"Could not connect to the server at {url}.")

        return response

    def _resolve_server_response(self, response: requests.Response) -> None:
        """
        Check the response from the server and raise errors as appropriate.
        """
        if response.status_code >= 400:
            message = response.json().get("detail")
            # print(response.text)
            if "Authorization" in message:
                print(message)
                message = "Please provide an Expected Parrot API key."
            raise CoopServerResponseError(message)

    def _json_handle_none(self, value: Any) -> Any:
        """
        Handle None values during JSON serialization.
        - Return "null" if the value is None. Otherwise, don't return anything.
        """
        if value is None:
            return "null"

    def _resolve_uuid(
        self, uuid: Union[str, UUID] = None, url: str = None
    ) -> Union[str, UUID]:
        """
        Resolve the uuid from a uuid or a url.
        """
        if not url and not uuid:
            raise CoopNoUUIDError("No uuid or url provided for the object.")
        if not uuid and url:
            uuid = url.split("/")[-1]
        return uuid

    @property
    def edsl_settings(self) -> dict:
        """
        Retrieve and return the EDSL settings stored on Coop.
        If no response is received within 5 seconds, return an empty dict.
        """
        from requests.exceptions import Timeout

        try:
            response = self._send_server_request(
                uri="api/v0/edsl-settings", method="GET", timeout=5
            )
            self._resolve_server_response(response)
            return response.json()
        except Timeout:
            return {}

    ################
    # Objects
    ################
    def create(
        self,
        object: EDSLObject,
        description: Optional[str] = None,
        visibility: Optional[VisibilityType] = "unlisted",
    ) -> dict:
        """
        Create an EDSL object in the Coop server.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(object)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="POST",
            payload={
                "description": description,
                "json_string": json.dumps(
                    object.to_dict(),
                    default=self._json_handle_none,
                ),
                "object_type": object_type,
                "visibility": visibility,
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "description": response_json.get("description"),
            "object_type": object_type,
            "url": f"{self.url}/content/{response_json.get('uuid')}",
            "uuid": response_json.get("uuid"),
            "version": self._edsl_version,
            "visibility": response_json.get("visibility"),
        }

    def get(
        self,
        uuid: Union[str, UUID] = None,
        url: str = None,
        expected_object_type: Optional[ObjectType] = None,
    ) -> EDSLObject:
        """
        Retrieve an EDSL object by its uuid or its url.
        - If the object's visibility is private, the user must be the owner.
        - Optionally, check if the retrieved object is of a certain type.

        :param uuid: the uuid of the object either in str or UUID format.
        :param url: the url of the object.
        :param expected_object_type: the expected type of the object.

        :return: the object instance.
        """
        uuid = self._resolve_uuid(uuid, url)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="GET",
            params={"uuid": uuid},
        )
        self._resolve_server_response(response)
        json_string = response.json().get("json_string")
        object_type = response.json().get("object_type")
        if expected_object_type and object_type != expected_object_type:
            raise Exception(f"Expected {expected_object_type=} but got {object_type=}")
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        object = edsl_class.from_dict(json.loads(json_string))
        return object

    def get_all(self, object_type: ObjectType) -> list[dict[str, Any]]:
        """
        Retrieve all objects of a certain type associated with the user.
        """
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        response = self._send_server_request(
            uri=f"api/v0/objects",
            method="GET",
            params={"type": object_type},
        )
        self._resolve_server_response(response)
        objects = [
            {
                "object": edsl_class.from_dict(json.loads(o.get("json_string"))),
                "uuid": o.get("uuid"),
                "version": o.get("version"),
                "description": o.get("description"),
                "visibility": o.get("visibility"),
                "url": f"{self.url}/content/{o.get('uuid')}",
            }
            for o in response.json()
        ]
        return objects

    def delete(self, uuid: Union[str, UUID] = None, url: str = None) -> dict:
        """
        Delete an object from the server.
        """
        uuid = self._resolve_uuid(uuid, url)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="DELETE",
            params={"uuid": uuid},
        )
        self._resolve_server_response(response)
        return response.json()

    def patch(
        self,
        uuid: Union[str, UUID] = None,
        url: str = None,
        description: Optional[str] = None,
        value: Optional[EDSLObject] = None,
        visibility: Optional[VisibilityType] = None,
    ) -> dict:
        """
        Change the attributes of an uploaded object
        - Only supports visibility for now
        """
        if description is None and visibility is None and value is None:
            raise Exception("Nothing to patch.")
        uuid = self._resolve_uuid(uuid, url)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="PATCH",
            params={"uuid": uuid},
            payload={
                "description": description,
                "json_string": (
                    json.dumps(
                        value.to_dict(),
                        default=self._json_handle_none,
                    )
                    if value
                    else None
                ),
                "visibility": visibility,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # Remote Cache
    ################
    def remote_cache_create(
        self,
        cache_entry: CacheEntry,
        visibility: VisibilityType = "private",
        description: Optional[str] = None,
    ) -> dict:
        """
        Create a single remote cache entry.
        If an entry with the same key already exists in the database, update it instead.

        :param cache_entry: The cache entry to send to the server.
        :param visibility: The visibility of the cache entry.
        :param optional description: A description for this entry in the remote cache.

        >>> entry = CacheEntry.example()
        >>> coop.remote_cache_create(cache_entry=entry)
        {'status': 'success', 'created_entry_count': 1, 'updated_entry_count': 0}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache",
            method="POST",
            payload={
                "json_string": json.dumps(cache_entry.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
                "description": description,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        created_entry_count = response_json.get("created_entry_count", 0)
        if created_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Upload new cache entries to server",
                cache_entry_count=created_entry_count,
            )
        return response.json()

    def remote_cache_create_many(
        self,
        cache_entries: list[CacheEntry],
        visibility: VisibilityType = "private",
        description: Optional[str] = None,
    ) -> dict:
        """
        Create many remote cache entries.
        If an entry with the same key already exists in the database, update it instead.

        :param cache_entries: The list of cache entries to send to the server.
        :param visibility: The visibility of the cache entries.
        :param optional description: A description for these entries in the remote cache.

        >>> entries = [CacheEntry.example(randomize=True) for _ in range(10)]
        >>> coop.remote_cache_create_many(cache_entries=entries)
        {'status': 'success', 'created_entry_count': 10, 'updated_entry_count': 0}
        """
        payload = [
            {
                "json_string": json.dumps(c.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
                "description": description,
            }
            for c in cache_entries
        ]
        response = self._send_server_request(
            uri="api/v0/remote-cache/many",
            method="POST",
            payload=payload,
        )
        self._resolve_server_response(response)
        response_json = response.json()
        created_entry_count = response_json.get("created_entry_count", 0)
        if created_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Upload new cache entries to server",
                cache_entry_count=created_entry_count,
            )
        return response.json()

    def remote_cache_get(
        self,
        exclude_keys: Optional[list[str]] = None,
    ) -> list[CacheEntry]:
        """
        Get all remote cache entries.

        :param optional exclude_keys: Exclude CacheEntry objects with these keys.

        >>> coop.remote_cache_get()
        [CacheEntry(...), CacheEntry(...), ...]
        """
        if exclude_keys is None:
            exclude_keys = []
        response = self._send_server_request(
            uri="api/v0/remote-cache/get-many",
            method="POST",
            payload={"keys": exclude_keys},
        )
        self._resolve_server_response(response)
        return [
            CacheEntry.from_dict(json.loads(v.get("json_string")))
            for v in response.json()
        ]

    def remote_cache_get_diff(
        self,
        client_cacheentry_keys: list[str],
    ) -> dict:
        """
        Get the difference between local and remote cache entries for a user.
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache/get-diff",
            method="POST",
            payload={"keys": client_cacheentry_keys},
        )
        self._resolve_server_response(response)
        response_json = response.json()
        response_dict = {
            "client_missing_cacheentries": [
                CacheEntry.from_dict(json.loads(c.get("json_string")))
                for c in response_json.get("client_missing_cacheentries", [])
            ],
            "server_missing_cacheentry_keys": response_json.get(
                "server_missing_cacheentry_keys", []
            ),
        }
        downloaded_entry_count = len(response_dict["client_missing_cacheentries"])
        if downloaded_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Download missing cache entries to client",
                cache_entry_count=downloaded_entry_count,
            )
        return response_dict

    def remote_cache_clear(self) -> dict:
        """
        Clear all remote cache entries.

        >>> entries = [CacheEntry.example(randomize=True) for _ in range(10)]
        >>> coop.remote_cache_create_many(cache_entries=entries)
        >>> coop.remote_cache_clear()
        {'status': 'success', 'deleted_entry_count': 10}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache/delete-all",
            method="DELETE",
        )
        self._resolve_server_response(response)
        response_json = response.json()
        deleted_entry_count = response_json.get("deleted_entry_count", 0)
        if deleted_entry_count > 0:
            self.remote_cache_create_log(
                response,
                description="Clear cache entries",
                cache_entry_count=deleted_entry_count,
            )
        return response.json()

    def remote_cache_create_log(
        self, response: requests.Response, description: str, cache_entry_count: int
    ) -> Union[dict, None]:
        """
        If a remote cache action has been completed successfully,
        log the action.
        """
        if 200 <= response.status_code < 300:
            log_response = self._send_server_request(
                uri="api/v0/remote-cache-log",
                method="POST",
                payload={
                    "description": description,
                    "cache_entry_count": cache_entry_count,
                },
            )
            self._resolve_server_response(log_response)
            return response.json()

    def remote_cache_clear_log(self) -> dict:
        """
        Clear all remote cache log entries.

        >>> coop.remote_cache_clear_log()
        {'status': 'success'}
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache-log/delete-all",
            method="DELETE",
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # Remote Inference
    ################
    def remote_inference_create(
        self,
        job: Jobs,
        description: Optional[str] = None,
        status: RemoteJobStatus = "queued",
        visibility: Optional[VisibilityType] = "unlisted",
        iterations: Optional[int] = 1,
    ) -> dict:
        """
        Send a remote inference job to the server.

        :param job: The EDSL job to send to the server.
        :param optional description: A description for this entry in the remote cache.
        :param status: The status of the job. Should be 'queued', unless you are debugging.
        :param visibility: The visibility of the cache entry.
        :param iterations: The number of times to run each interview.

        >>> job = Jobs.example()
        >>> coop.remote_inference_create(job=job, description="My job")
        {'uuid': '9f8484ee-b407-40e4-9652-4133a7236c9c', 'description': 'My job', 'status': 'queued', 'visibility': 'unlisted', 'version': '0.1.29.dev4'}
        """
        response = self._send_server_request(
            uri="api/v0/remote-inference",
            method="POST",
            payload={
                "json_string": json.dumps(
                    job.to_dict(),
                    default=self._json_handle_none,
                ),
                "description": description,
                "status": status,
                "iterations": iterations,
                "visibility": visibility,
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "uuid": response_json.get("job_uuid"),
            "description": response_json.get("description"),
            "status": response_json.get("status"),
            "iterations": response_json.get("iterations"),
            "visibility": response_json.get("visibility"),
            "version": self._edsl_version,
        }

    def remote_inference_get(
        self, job_uuid: Optional[str] = None, results_uuid: Optional[str] = None
    ) -> dict:
        """
        Get the details of a remote inference job.
        You can pass either the job uuid or the results uuid as a parameter.
        If you pass both, the job uuid will be prioritized.

        :param job_uuid: The UUID of the EDSL job.
        :param results_uuid: The UUID of the results associated with the EDSL job.

        >>> coop.remote_inference_get("9f8484ee-b407-40e4-9652-4133a7236c9c")
        {'jobs_uuid': '9f8484ee-b407-40e4-9652-4133a7236c9c', 'results_uuid': 'dd708234-31bf-4fe1-8747-6e232625e026', 'results_url': 'https://www.expectedparrot.com/content/dd708234-31bf-4fe1-8747-6e232625e026', 'status': 'completed', 'reason': None, 'price': 16, 'version': '0.1.29.dev4'}
        """
        if job_uuid is None and results_uuid is None:
            raise ValueError("Either job_uuid or results_uuid must be provided.")
        elif job_uuid is not None:
            params = {"job_uuid": job_uuid}
        else:
            params = {"results_uuid": results_uuid}

        response = self._send_server_request(
            uri="api/v0/remote-inference",
            method="GET",
            params=params,
        )
        self._resolve_server_response(response)
        data = response.json()
        return {
            "job_uuid": data.get("job_uuid"),
            "results_uuid": data.get("results_uuid"),
            "results_url": f"{self.url}/content/{data.get('results_uuid')}",
            "status": data.get("status"),
            "reason": data.get("reason"),
            "credits_consumed": data.get("price"),
            "version": data.get("version"),
        }

    def remote_inference_cost(self, input: Union[Jobs, Survey]) -> int:
        """
        Get the cost of a remote inference job.

        :param input: The EDSL job to send to the server.

        >>> job = Jobs.example()
        >>> coop.remote_inference_cost(input=job)
        16
        """
        if isinstance(input, Jobs):
            job = input
        elif isinstance(input, Survey):
            job = Jobs(survey=input)
        else:
            raise TypeError("Input must be either a Job or a Survey.")

        response = self._send_server_request(
            uri="api/v0/remote-inference/cost",
            method="POST",
            payload={
                "json_string": json.dumps(
                    job.to_dict(),
                    default=self._json_handle_none,
                ),
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "credits": response_json.get("cost_in_credits"),
            "usd": response_json.get("cost_in_usd"),
        }

    ################
    # Remote Errors
    ################
    def error_create(self, error_data: str) -> dict:
        """
        Send an error message to the server.
        """
        response = self._send_server_request(
            uri="api/v0/errors",
            method="POST",
            payload={
                "json_string": json.dumps(error_data),
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # DUNDER METHODS
    ################
    def __repr__(self):
        """Return a string representation of the client."""
        return f"Client(api_key='{self.api_key}', url='{self.url}')"

    ################
    # EXPERIMENTAL
    ################
    async def remote_async_execute_model_call(
        self, model_dict: dict, user_prompt: str, system_prompt: str
    ) -> dict:
        url = self.url + "/inference/"
        # print("Now using url: ", url)
        data = {
            "model_dict": model_dict,
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }
        # Use aiohttp to send a POST request asynchronously
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=data) as response:
                response_data = await response.json()
        return response_data

    def web(
        self,
        survey: dict,
        platform: Literal[
            "google_forms", "lime_survey", "survey_monkey"
        ] = "lime_survey",
        email=None,
    ):
        url = f"{self.url}/api/v0/export_to_{platform}"
        if email:
            data = {"json_string": json.dumps({"survey": survey, "email": email})}
        else:
            data = {"json_string": json.dumps({"survey": survey, "email": ""})}

        response_json = requests.post(url, headers=self.headers, data=json.dumps(data))

        return response_json

    def fetch_prices(self) -> dict:
        """
        Fetch model prices from Coop. If the request fails, return an empty dict.
        """

        from edsl.coop.PriceFetcher import PriceFetcher

        from edsl.config import CONFIG

        if bool(CONFIG.get("EDSL_FETCH_TOKEN_PRICES")):
            price_fetcher = PriceFetcher()
            return price_fetcher.fetch_prices()
        else:
            return {}

    def fetch_rate_limit_config_vars(self) -> dict:
        """
        Fetch a dict of rate limit config vars from Coop.

        The dict keys are RPM and TPM variables like EDSL_SERVICE_RPM_OPENAI.
        """
        response = self._send_server_request(
            uri="api/v0/config-vars",
            method="GET",
        )
        self._resolve_server_response(response)
        data = response.json()
        return data


if __name__ == "__main__":
    sheet_data = fetch_sheet_data()
    if sheet_data:
        print(f"Successfully fetched {len(sheet_data)} rows of data.")
        print("First row:", sheet_data[0])
    else:
        print("Failed to fetch sheet data.")


def main():
    """
    A simple example for the coop client
    """
    from uuid import uuid4
    from edsl import (
        Agent,
        AgentList,
        Cache,
        Notebook,
        QuestionFreeText,
        QuestionMultipleChoice,
        Results,
        Scenario,
        ScenarioList,
        Survey,
    )
    from edsl.coop import Coop
    from edsl.data.CacheEntry import CacheEntry
    from edsl.jobs import Jobs

    # init & basics
    API_KEY = "b"
    coop = Coop(api_key=API_KEY)
    coop
    coop.edsl_settings

    ##############
    # A. A simple example
    ##############
    # .. create and manipulate an object through the Coop client
    response = coop.create(QuestionMultipleChoice.example())
    coop.get(uuid=response.get("uuid"))
    coop.get(uuid=response.get("uuid"), expected_object_type="question")
    coop.get(url=response.get("url"))
    coop.create(QuestionMultipleChoice.example())
    coop.get_all("question")
    coop.patch(uuid=response.get("uuid"), visibility="private")
    coop.patch(uuid=response.get("uuid"), description="hey")
    coop.patch(uuid=response.get("uuid"), value=QuestionFreeText.example())
    # coop.patch(uuid=response.get("uuid"), value=Survey.example()) - should throw error
    coop.get(uuid=response.get("uuid"))
    coop.delete(uuid=response.get("uuid"))

    # .. create and manipulate an object through the class
    response = QuestionMultipleChoice.example().push()
    QuestionMultipleChoice.pull(uuid=response.get("uuid"))
    QuestionMultipleChoice.pull(url=response.get("url"))
    QuestionMultipleChoice.patch(uuid=response.get("uuid"), visibility="private")
    QuestionMultipleChoice.patch(uuid=response.get("uuid"), description="hey")
    QuestionMultipleChoice.patch(
        uuid=response.get("uuid"), value=QuestionFreeText.example()
    )
    QuestionMultipleChoice.pull(response.get("uuid"))
    QuestionMultipleChoice.delete(response.get("uuid"))

    ##############
    # B. Examples with all objects
    ##############
    OBJECTS = [
        ("agent", Agent),
        ("agent_list", AgentList),
        ("cache", Cache),
        ("notebook", Notebook),
        ("question", QuestionMultipleChoice),
        ("results", Results),
        ("scenario", Scenario),
        ("scenario_list", ScenarioList),
        ("survey", Survey),
    ]
    for object_type, cls in OBJECTS:
        print(f"Testing {object_type} objects")
        # 1. Delete existing objects
        existing_objects = coop.get_all(object_type)
        for item in existing_objects:
            coop.delete(uuid=item.get("uuid"))
        # 2. Create new objects
        example = cls.example()
        response_1 = coop.create(example)
        response_2 = coop.create(cls.example(), visibility="private")
        response_3 = coop.create(cls.example(), visibility="public")
        response_4 = coop.create(
            cls.example(), visibility="unlisted", description="hey"
        )
        # 3. Retrieve all objects
        objects = coop.get_all(object_type)
        assert len(objects) == 4
        # 4. Try to retrieve an item that does not exist
        try:
            coop.get(uuid=uuid4())
        except Exception as e:
            print(e)
        # 5. Try to retrieve all test objects by their uuids
        for response in [response_1, response_2, response_3, response_4]:
            coop.get(uuid=response.get("uuid"))
        # 6. Change visibility of all objects
        for item in objects:
            coop.patch(uuid=item.get("uuid"), visibility="private")
        # 6. Change description of all objects
        for item in objects:
            coop.patch(uuid=item.get("uuid"), description="hey")
        # 7. Delete all objects
        for item in objects:
            coop.delete(uuid=item.get("uuid"))
        assert len(coop.get_all(object_type)) == 0

    ##############
    # C. Remote Cache
    ##############
    # clear
    coop.remote_cache_clear()
    assert coop.remote_cache_get() == []
    # create one remote cache entry
    cache_entry = CacheEntry.example()
    cache_entry.to_dict()
    coop.remote_cache_create(cache_entry)
    # create many remote cache entries
    cache_entries = [CacheEntry.example(randomize=True) for _ in range(10)]
    coop.remote_cache_create_many(cache_entries)
    # get all remote cache entries
    coop.remote_cache_get()
    coop.remote_cache_get(exclude_keys=[])
    coop.remote_cache_get(exclude_keys=["a"])
    exclude_keys = [cache_entry.key for cache_entry in cache_entries]
    coop.remote_cache_get(exclude_keys)
    # clear
    coop.remote_cache_clear()
    coop.remote_cache_get()

    ##############
    # D. Remote Inference
    ##############
    job = Jobs.example()
    coop.remote_inference_cost(job)
    job_coop_object = coop.remote_inference_create(job)
    job_coop_results = coop.remote_inference_get(job_coop_object.get("uuid"))
    coop.get(uuid=job_coop_results.get("results_uuid"))

    ##############
    # E. Errors
    ##############
    coop.error_create({"something": "This is an error message"})
    coop.api_key = None
    coop.error_create({"something": "This is an error message"})



from __future__ import annotations
from typing import Any, Optional
from uuid import uuid4

from pydantic import field_validator

from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.ResponseValidatorABC import ResponseValidatorABC

from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions.decorators import inject_exception

from pydantic import BaseModel
from typing import Optional, Any, List

from edsl.exceptions import QuestionAnswerValidationError


class FreeTextResponse(BaseModel):
    """
    Validator for free text response questions.
    """

    answer: str
    generated_tokens: Optional[str] = None


class FreeTextResponseValidator(ResponseValidatorABC):
    required_params = []
    valid_examples = [({"answer": "This is great"}, {})]
    invalid_examples = [
        (
            {"answer": None},
            {},
            "Answer code must not be missing.",
        ),
    ]

    def fix(self, response, verbose=False):
        return {
            "answer": str(response.get("generated_tokens")),
            "generated_tokens": str(response.get("generated_tokens")),
        }


class QuestionFreeText(QuestionBase):
    """This question prompts the agent to respond with free text."""

    question_type = "free_text"
    _response_model = FreeTextResponse
    response_validator_class = FreeTextResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        answering_instructions: Optional[str] = None,
        question_presentation: Optional[str] = None,
    ):
        """Instantiate a new QuestionFreeText.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.answering_instructions = answering_instructions
        self.question_presentation = question_presentation

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <div>
        <textarea id="{{ question_name }}" name="{{ question_name }}"></textarea>
        </div>
        """
        ).render(question_name=self.question_name)
        return question_html_content

    @classmethod
    @inject_exception
    def example(cls, randomize: bool = False) -> QuestionFreeText:
        """Return an example instance of a free text question."""
        addition = "" if not randomize else str(uuid4())
        return cls(question_name="how_are_you", question_text=f"How are you?{addition}")


def main():
    """Create an example question and demonstrate its functionality."""
    from edsl.questions.QuestionFreeText import QuestionFreeText

    q = QuestionFreeText.example()
    q.question_text
    q.question_name
    # q.instructions
    # validate an answer
    q._validate_answer({"answer": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer({"answer"})
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl import QuestionList, Scenario, Model


def summarize(texts, seed_phrase, n_bullets, n_words, models=None):
    if models is None:
        models = Model()
    s = Scenario(
        text=texts, seed_phrase=seed_phrase, n_bullets=n_bullets, n_words=n_words
    ).expand("text")
    QuestionList(
        question_text="""
        I have the following TEXT EXAMPLE :
        {{ text_example_json }}
        Please summarize the main point of this EXAMPLE {{seed_phrase }} into {{ n_bullets }} bullet points, where
        each bullet point is a {{ n_words }} word phrase.
        """,
        question_name="summarize",
    ).by(s).by(models).run()



from typing import Optional

from edsl import Model
from edsl.auto.StageQuestions import StageQuestions
from edsl.auto.StagePersona import StagePersona
from edsl.auto.StagePersonaDimensions import StagePersonaDimensions
from edsl.auto.StagePersonaDimensionValues import StagePersonaDimensionValues
from edsl.auto.StagePersonaDimensionValueRanges import (
    StagePersonaDimensionValueRanges,
)
from edsl.auto.StageLabelQuestions import StageLabelQuestions
from edsl.auto.StageGenerateSurvey import StageGenerateSurvey

# from edsl.auto.StageBase import gen_pipeline

from edsl.auto.utilities import agent_generator, create_agents, gen_pipeline


class AutoStudy:
    def __init__(
        self,
        overall_question: str,
        population: str,
        model: Optional["Model"] = None,
        survey: Optional["Survey"] = None,
        agent_list: Optional["AgentList"] = None,
        default_num_agents=11,
    ):
        self.overall_question = overall_question
        self.population = population
        self._survey = survey
        self._agent_list = agent_list
        self._agent_list_generator = None
        self._persona_mapping = None
        self._results = None
        self.default_num_agents = default_num_agents
        self.model = model or Model()

    @property
    def survey(self):
        if self._survey is None:
            self._survey = self._create_survey()
        return self._survey

    @property
    def persona_mapping(self):
        if self._persona_mapping is None:
            self._persona_mapping = self._create_persona_mapping()
        return self._persona_mapping

    @property
    def agent_list_generator(self):
        if self._agent_list_generator is None:
            self._agent_list_generator = self._create_agent_list_generator()
        return self._agent_list_generator

    @property
    def results(self):
        if self._results is None:
            self._results = self._create_results()
        return self._results

    def _create_survey(self):
        survey_pipline_stages = [
            StageQuestions,
            StageLabelQuestions,
            StageGenerateSurvey,
        ]
        survey_pipeline = gen_pipeline(survey_pipline_stages)
        return survey_pipeline.process(
            data=survey_pipeline.input(
                overall_question=self.overall_question, population=self.population
            )
        ).survey

    def _create_persona_mapping(self):
        persona_pipeline_stages = [
            StageQuestions,
            StagePersona,
            StagePersonaDimensions,
            StagePersonaDimensionValues,
            StagePersonaDimensionValueRanges,
        ]

        persona_pipeline = gen_pipeline(persona_pipeline_stages)
        sample_agent_results = persona_pipeline.process(
            persona_pipeline.input(
                overall_question=overall_question, population=self.population
            )
        )
        return sample_agent_results

    def _create_agent_list_generator(self):
        return agent_generator(
            persona=self.persona_mapping.persona,
            dimension_dict=self.persona_mapping.mapping,
        )

    def agent_list(self, num_agents):
        return create_agents(
            agent_generator=self.agent_list_generator,
            survey=self.survey,
            num_agents=num_agents,
        )

    def _create_results(self, num_agents=None):
        if num_agents is None:
            num_agents = self.default_num_agents
        agent_list = self.agent_list(num_agents)
        return self.survey.by(agent_list).by(self.model).run()


if __name__ == "__main__":
    overall_question = "Should online platforms be regulated with respect to selling electric scooters?"
    auto_study = AutoStudy(overall_question, population="US Adults")

    results = auto_study.results



"""This module contains the Interview class, which is responsible for conducting an interview asynchronously."""

from __future__ import annotations
import asyncio
from typing import Any, Type, List, Generator, Optional, Union
import copy

from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
    RetryError,
)

from edsl import CONFIG
from edsl.surveys.base import EndOfSurvey
from edsl.exceptions import QuestionAnswerValidationError
from edsl.exceptions import QuestionAnswerValidationError
from edsl.data_transfer_models import AgentResponseDict, EDSLResultObjectInput

from edsl.jobs.buckets.ModelBuckets import ModelBuckets
from edsl.jobs.Answers import Answers
from edsl.jobs.tasks.QuestionTaskCreator import QuestionTaskCreator
from edsl.jobs.tasks.TaskCreators import TaskCreators
from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog
from edsl.jobs.interviews.InterviewExceptionCollection import (
    InterviewExceptionCollection,
)

# from edsl.jobs.interviews.InterviewStatusMixin import InterviewStatusMixin

from edsl.surveys.base import EndOfSurvey
from edsl.jobs.buckets.ModelBuckets import ModelBuckets
from edsl.jobs.interviews.InterviewExceptionEntry import InterviewExceptionEntry
from edsl.jobs.tasks.task_status_enum import TaskStatus
from edsl.jobs.tasks.QuestionTaskCreator import QuestionTaskCreator


from edsl import Agent, Survey, Scenario, Cache
from edsl.language_models import LanguageModel
from edsl.questions import QuestionBase
from edsl.agents.InvigilatorBase import InvigilatorBase

from edsl.exceptions.language_models import LanguageModelNoResponseError

from edsl.jobs.interviews.InterviewStatusLog import InterviewStatusLog
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary


from edsl import CONFIG

EDSL_BACKOFF_START_SEC = float(CONFIG.get("EDSL_BACKOFF_START_SEC"))
EDSL_BACKOFF_MAX_SEC = float(CONFIG.get("EDSL_BACKOFF_MAX_SEC"))
EDSL_MAX_ATTEMPTS = int(CONFIG.get("EDSL_MAX_ATTEMPTS"))


class Interview:
    """
    An 'interview' is one agent answering one survey, with one language model, for a given scenario.

    The main method is `async_conduct_interview`, which conducts the interview asynchronously.
    Most of the class is dedicated to creating the tasks for each question in the survey, and then running them.
    """

    def __init__(
        self,
        agent: Agent,
        survey: Survey,
        scenario: Scenario,
        model: Type["LanguageModel"],
        debug: Optional[bool] = False,
        iteration: int = 0,
        cache: Optional["Cache"] = None,
        sidecar_model: Optional["LanguageModel"] = None,
        skip_retry: bool = False,
        raise_validation_errors: bool = True,
    ):
        """Initialize the Interview instance.

        :param agent: the agent being interviewed.
        :param survey: the survey being administered to the agent.
        :param scenario: the scenario that populates the survey questions.
        :param model: the language model used to answer the questions.
        :param debug: if True, run without calls to the language model.
        :param iteration: the iteration number of the interview.
        :param cache: the cache used to store the answers.
        :param sidecar_model: a sidecar model used to answer questions.

        >>> i = Interview.example()
        >>> i.task_creators
        {}

        >>> i.exceptions
        {}

        >>> _ = asyncio.run(i.async_conduct_interview())
        >>> i.task_status_logs['q0']
        [{'log_time': ..., 'value': <TaskStatus.NOT_STARTED: 1>}, {'log_time': ..., 'value': <TaskStatus.WAITING_FOR_DEPENDENCIES: 2>}, {'log_time': ..., 'value': <TaskStatus.API_CALL_IN_PROGRESS: 7>}, {'log_time': ..., 'value': <TaskStatus.SUCCESS: 8>}]

        >>> i.to_index
        {'q0': 0, 'q1': 1, 'q2': 2}

        """
        self.agent = agent
        self.survey = copy.deepcopy(survey)
        self.scenario = scenario
        self.model = model
        self.debug = debug
        self.iteration = iteration
        self.cache = cache
        self.answers: dict[
            str, str
        ] = Answers()  # will get filled in as interview progresses
        self.sidecar_model = sidecar_model

        # Trackers
        self.task_creators = TaskCreators()  # tracks the task creators
        self.exceptions = InterviewExceptionCollection()

        self._task_status_log_dict = InterviewStatusLog()
        self.skip_retry = skip_retry
        self.raise_validation_errors = raise_validation_errors

        # dictionary mapping question names to their index in the survey.
        self.to_index = {
            question_name: index
            for index, question_name in enumerate(self.survey.question_names)
        }

        self.failed_questions = []

    @property
    def has_exceptions(self) -> bool:
        """Return True if there are exceptions."""
        return len(self.exceptions) > 0

    @property
    def task_status_logs(self) -> InterviewStatusLog:
        """Return the task status logs for the interview.

        The keys are the question names; the values are the lists of status log changes for each task.
        """
        for task_creator in self.task_creators.values():
            self._task_status_log_dict[
                task_creator.question.question_name
            ] = task_creator.status_log
        return self._task_status_log_dict

    @property
    def token_usage(self) -> InterviewTokenUsage:
        """Determine how many tokens were used for the interview."""
        return self.task_creators.token_usage

    @property
    def interview_status(self) -> InterviewStatusDictionary:
        """Return a dictionary mapping task status codes to counts."""
        return self.task_creators.interview_status

    # region: Serialization
    def _to_dict(self, include_exceptions=True) -> dict[str, Any]:
        """Return a dictionary representation of the Interview instance.
        This is just for hashing purposes.

        >>> i = Interview.example()
        >>> hash(i)
        1217840301076717434
        """
        d = {
            "agent": self.agent._to_dict(),
            "survey": self.survey._to_dict(),
            "scenario": self.scenario._to_dict(),
            "model": self.model._to_dict(),
            "iteration": self.iteration,
            "exceptions": {},
        }
        if include_exceptions:
            d["exceptions"] = self.exceptions.to_dict()
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "Interview":
        """Return an Interview instance from a dictionary."""
        agent = Agent.from_dict(d["agent"])
        survey = Survey.from_dict(d["survey"])
        scenario = Scenario.from_dict(d["scenario"])
        model = LanguageModel.from_dict(d["model"])
        iteration = d["iteration"]
        interview = cls(
            agent=agent,
            survey=survey,
            scenario=scenario,
            model=model,
            iteration=iteration,
        )
        if "exceptions" in d:
            exceptions = InterviewExceptionCollection.from_dict(d["exceptions"])
            interview.exceptions = exceptions
        return interview

    def __hash__(self) -> int:
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self._to_dict(include_exceptions=False))

    def __eq__(self, other: "Interview") -> bool:
        """
        >>> from edsl.jobs.interviews.Interview import Interview; i = Interview.example(); d = i._to_dict(); i2 = Interview.from_dict(d); i == i2
        True
        """
        return hash(self) == hash(other)

    # endregion

    # region: Creating tasks
    @property
    def dag(self) -> "DAG":
        """Return the directed acyclic graph for the survey.

        The DAG, or directed acyclic graph, is a dictionary that maps question names to their dependencies.
        It is used to determine the order in which questions should be answered.
        This reflects both agent 'memory' considerations and 'skip' logic.
        The 'textify' parameter is set to True, so that the question names are returned as strings rather than integer indices.

        >>> i = Interview.example()
        >>> i.dag == {'q2': {'q0'}, 'q1': {'q0'}}
        True
        """
        return self.survey.dag(textify=True)

    def _build_question_tasks(
        self,
        model_buckets: ModelBuckets,
    ) -> list[asyncio.Task]:
        """Create a task for each question, with dependencies on the questions that must be answered before this one can be answered.

        :param debug: whether to use debug mode, in which case `InvigilatorDebug` is used.
        :param model_buckets: the model buckets used to track and control usage rates.
        """
        tasks = []
        for question in self.survey.questions:
            tasks_that_must_be_completed_before = list(
                self._get_tasks_that_must_be_completed_before(
                    tasks=tasks, question=question
                )
            )
            question_task = self._create_question_task(
                question=question,
                tasks_that_must_be_completed_before=tasks_that_must_be_completed_before,
                model_buckets=model_buckets,
                iteration=self.iteration,
            )
            tasks.append(question_task)
        return tuple(tasks)

    def _get_tasks_that_must_be_completed_before(
        self, *, tasks: list[asyncio.Task], question: "QuestionBase"
    ) -> Generator[asyncio.Task, None, None]:
        """Return the tasks that must be completed before the given question can be answered.

        :param tasks: a list of tasks that have been created so far.
        :param question: the question for which we are determining dependencies.

        If a question has no dependencies, this will be an empty list, [].
        """
        parents_of_focal_question = self.dag.get(question.question_name, [])
        for parent_question_name in parents_of_focal_question:
            yield tasks[self.to_index[parent_question_name]]

    def _create_question_task(
        self,
        *,
        question: QuestionBase,
        tasks_that_must_be_completed_before: list[asyncio.Task],
        model_buckets: ModelBuckets,
        iteration: int = 0,
    ) -> asyncio.Task:
        """Create a task that depends on the passed-in dependencies that are awaited before the task is run.

        :param question: the question to be answered. This is the question we are creating a task for.
        :param tasks_that_must_be_completed_before: the tasks that must be completed before the focal task is run.
        :param model_buckets: the model buckets used to track and control usage rates.
        :param debug: whether to use debug mode, in which case `InvigilatorDebug` is used.
        :param iteration: the iteration number for the interview.

        The task is created by a `QuestionTaskCreator`, which is responsible for creating the task and managing its dependencies.
        It is passed a reference to the function that will be called to answer the question.
        It is passed a list "tasks_that_must_be_completed_before" that are awaited before the task is run.
        These are added as a dependency to the focal task.
        """
        task_creator = QuestionTaskCreator(
            question=question,
            answer_question_func=self._answer_question_and_record_task,
            token_estimator=self._get_estimated_request_tokens,
            model_buckets=model_buckets,
            iteration=iteration,
        )
        for task in tasks_that_must_be_completed_before:
            task_creator.add_dependency(task)

        self.task_creators.update(
            {question.question_name: task_creator}
        )  # track this task creator
        return task_creator.generate_task()

    def _get_estimated_request_tokens(self, question) -> float:
        """Estimate the number of tokens that will be required to run the focal task."""
        from edsl.scenarios.FileStore import FileStore

        invigilator = self._get_invigilator(question=question)
        # TODO: There should be a way to get a more accurate estimate.
        combined_text = ""
        file_tokens = 0
        for prompt in invigilator.get_prompts().values():
            if hasattr(prompt, "text"):
                combined_text += prompt.text
            elif isinstance(prompt, str):
                combined_text += prompt
            elif isinstance(prompt, list):
                for file in prompt:
                    if isinstance(file, FileStore):
                        file_tokens += file.size * 0.25
            else:
                raise ValueError(f"Prompt is of type {type(prompt)}")
        return len(combined_text) / 4.0 + file_tokens

    async def _answer_question_and_record_task(
        self,
        *,
        question: "QuestionBase",
        task=None,
    ) -> "AgentResponseDict":
        """Answer a question and records the task."""

        had_language_model_no_response_error = False

        @retry(
            stop=stop_after_attempt(EDSL_MAX_ATTEMPTS),
            wait=wait_exponential(
                multiplier=EDSL_BACKOFF_START_SEC, max=EDSL_BACKOFF_MAX_SEC
            ),
            retry=retry_if_exception_type(LanguageModelNoResponseError),
            reraise=True,
        )
        async def attempt_answer():
            nonlocal had_language_model_no_response_error

            invigilator = self._get_invigilator(question)

            if self._skip_this_question(question):
                return invigilator.get_failed_task_result(
                    failure_reason="Question skipped."
                )

            try:
                response: EDSLResultObjectInput = (
                    await invigilator.async_answer_question()
                )
                if response.validated:
                    self.answers.add_answer(response=response, question=question)
                    self._cancel_skipped_questions(question)
                else:
                    # When a question is not validated, it is not added to the answers.
                    # this should also cancel and dependent children questions.
                    # Is that happening now?
                    if (
                        hasattr(response, "exception_occurred")
                        and response.exception_occurred
                    ):
                        raise response.exception_occurred

            except QuestionAnswerValidationError as e:
                self._handle_exception(e, invigilator, task)
                return invigilator.get_failed_task_result(
                    failure_reason="Question answer validation failed."
                )

            except asyncio.TimeoutError as e:
                self._handle_exception(e, invigilator, task)
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model timed out for question '{question.question_name}.'"
                )

            except Exception as e:
                self._handle_exception(e, invigilator, task)

            if "response" not in locals():
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model did not return a response for question '{question.question_name}.'"
                )

            # if it gets here, it means the no response error was fixed
            if (
                question.question_name in self.exceptions
                and had_language_model_no_response_error
            ):
                self.exceptions.record_fixed_question(question.question_name)

            return response

        try:
            return await attempt_answer()
        except RetryError as retry_error:
            # All retries have failed for LanguageModelNoResponseError
            original_error = retry_error.last_attempt.exception()
            self._handle_exception(
                original_error, self._get_invigilator(question), task
            )
            raise original_error  # Re-raise the original error after handling

    def _get_invigilator(self, question: QuestionBase) -> InvigilatorBase:
        """Return an invigilator for the given question.

        :param question: the question to be answered
        :param debug: whether to use debug mode, in which case `InvigilatorDebug` is used.
        """
        invigilator = self.agent.create_invigilator(
            question=question,
            scenario=self.scenario,
            model=self.model,
            debug=False,
            survey=self.survey,
            memory_plan=self.survey.memory_plan,
            current_answers=self.answers,
            iteration=self.iteration,
            cache=self.cache,
            sidecar_model=self.sidecar_model,
            raise_validation_errors=self.raise_validation_errors,
        )
        """Return an invigilator for the given question."""
        return invigilator

    def _skip_this_question(self, current_question: "QuestionBase") -> bool:
        """Determine if the current question should be skipped.

        :param current_question: the question to be answered.
        """
        current_question_index = self.to_index[current_question.question_name]

        answers = self.answers | self.scenario | self.agent["traits"]
        skip = self.survey.rule_collection.skip_question_before_running(
            current_question_index, answers
        )
        return skip

    def _handle_exception(
        self, e: Exception, invigilator: "InvigilatorBase", task=None
    ):
        import copy

        # breakpoint()

        answers = copy.copy(self.answers)
        exception_entry = InterviewExceptionEntry(
            exception=e,
            invigilator=invigilator,
            answers=answers,
        )
        if task:
            task.task_status = TaskStatus.FAILED
        self.exceptions.add(invigilator.question.question_name, exception_entry)

        if self.raise_validation_errors:
            if isinstance(e, QuestionAnswerValidationError):
                raise e

        if hasattr(self, "stop_on_exception"):
            stop_on_exception = self.stop_on_exception
        else:
            stop_on_exception = False

        if stop_on_exception:
            raise e

    def _cancel_skipped_questions(self, current_question: QuestionBase) -> None:
        """Cancel the tasks for questions that are skipped.

        :param current_question: the question that was just answered.

        It first determines the next question, given the current question and the current answers.
        If the next question is the end of the survey, it cancels all remaining tasks.
        If the next question is after the current question, it cancels all tasks between the current question and the next question.
        """
        current_question_index: int = self.to_index[current_question.question_name]

        next_question: Union[
            int, EndOfSurvey
        ] = self.survey.rule_collection.next_question(
            q_now=current_question_index,
            answers=self.answers | self.scenario | self.agent["traits"],
        )

        next_question_index = next_question.next_q

        def cancel_between(start, end):
            """Cancel the tasks between the start and end indices."""
            for i in range(start, end):
                self.tasks[i].cancel()

        if next_question_index == EndOfSurvey:
            cancel_between(current_question_index + 1, len(self.survey.questions))
            return

        if next_question_index > (current_question_index + 1):
            cancel_between(current_question_index + 1, next_question_index)

    # endregion

    # region: Conducting the interview
    async def async_conduct_interview(
        self,
        model_buckets: Optional[ModelBuckets] = None,
        stop_on_exception: bool = False,
        sidecar_model: Optional["LanguageModel"] = None,
        raise_validation_errors: bool = True,
    ) -> tuple["Answers", List[dict[str, Any]]]:
        """
        Conduct an Interview asynchronously.
        It returns a tuple with the answers and a list of valid results.

        :param model_buckets: a dictionary of token buckets for the model.
        :param debug: run without calls to LLM.
        :param stop_on_exception: if True, stops the interview if an exception is raised.
        :param sidecar_model: a sidecar model used to answer questions.

        Example usage:

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> result['q0']
        'yes'

        >>> i = Interview.example(throw_exception = True)
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> i.exceptions
        {'q0': ...
        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview(stop_on_exception = True))
        Traceback (most recent call last):
        ...
        asyncio.exceptions.CancelledError
        """
        self.sidecar_model = sidecar_model
        self.stop_on_exception = stop_on_exception

        # if no model bucket is passed, create an 'infinity' bucket with no rate limits
        if model_buckets is None or hasattr(self.agent, "answer_question_directly"):
            model_buckets = ModelBuckets.infinity_bucket()

        ## This is the key part---it creates a task for each question,
        ## with dependencies on the questions that must be answered before this one can be answered.
        self.tasks = self._build_question_tasks(model_buckets=model_buckets)

        ## 'Invigilators' are used to administer the survey
        self.invigilators = [
            self._get_invigilator(question) for question in self.survey.questions
        ]
        await asyncio.gather(
            *self.tasks, return_exceptions=not stop_on_exception
        )  # not stop_on_exception)
        self.answers.replace_missing_answers_with_none(self.survey)
        valid_results = list(self._extract_valid_results())
        return self.answers, valid_results

    # endregion

    # region: Extracting results and recording errors
    def _extract_valid_results(self) -> Generator["Answers", None, None]:
        """Extract the valid results from the list of results.

        It iterates through the tasks and invigilators, and yields the results of the tasks that are done.
        If a task is not done, it raises a ValueError.
        If an exception is raised in the task, it records the exception in the Interview instance except if the task was cancelled, which is expected behavior.

        >>> i = Interview.example()
        >>> result, _ = asyncio.run(i.async_conduct_interview())
        >>> results = list(i._extract_valid_results())
        >>> len(results) == len(i.survey)
        True
        """
        assert len(self.tasks) == len(self.invigilators)

        for task, invigilator in zip(self.tasks, self.invigilators):
            if not task.done():
                raise ValueError(f"Task {task.get_name()} is not done.")

            try:
                result = task.result()
            except asyncio.CancelledError as e:  # task was cancelled
                result = invigilator.get_failed_task_result(
                    failure_reason="Task was cancelled."
                )
            except Exception as e:  # any other kind of exception in the task
                result = invigilator.get_failed_task_result(
                    failure_reason=f"Task failed with exception: {str(e)}."
                )
                exception_entry = InterviewExceptionEntry(
                    exception=e,
                    invigilator=invigilator,
                )
                self.exceptions.add(task.get_name(), exception_entry)

            yield result

    # endregion

    # region: Magic methods
    def __repr__(self) -> str:
        """Return a string representation of the Interview instance."""
        return f"Interview(agent = {repr(self.agent)}, survey = {repr(self.survey)}, scenario = {repr(self.scenario)}, model = {repr(self.model)})"

    def duplicate(self, iteration: int, cache: "Cache") -> Interview:
        """Duplicate the interview, but with a new iteration number and cache.

        >>> i = Interview.example()
        >>> i2 = i.duplicate(1, None)
        >>> i.iteration + 1 == i2.iteration
        True

        """
        return Interview(
            agent=self.agent,
            survey=self.survey,
            scenario=self.scenario,
            model=self.model,
            iteration=iteration,
            cache=cache,
            skip_retry=self.skip_retry,
        )

    @classmethod
    def example(self, throw_exception: bool = False) -> Interview:
        """Return an example Interview instance."""
        from edsl.agents import Agent
        from edsl.surveys import Survey
        from edsl.scenarios import Scenario
        from edsl.language_models import LanguageModel

        def f(self, question, scenario):
            return "yes"

        agent = Agent.example()
        agent.add_direct_question_answering_method(f)
        survey = Survey.example()
        scenario = Scenario.example()
        model = LanguageModel.example()
        if throw_exception:
            model = LanguageModel.example(test_model=True, throw_exception=True)
            agent = Agent.example()
            return Interview(agent=agent, survey=survey, scenario=scenario, model=model)
        return Interview(agent=agent, survey=survey, scenario=scenario, model=model)


if __name__ == "__main__":
    import doctest

    # add ellipsis
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations

# from decimal import Decimal
from random import uniform
from typing import Any, Optional, Union, Literal

from pydantic import BaseModel, Field, field_validator

from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import NumericalOrNoneDescriptor
from edsl.questions.decorators import inject_exception
from edsl.questions.ResponseValidatorABC import ResponseValidatorABC
from edsl.exceptions.questions import QuestionAnswerValidationError


def create_numeric_response(
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
    permissive=False,
):
    field_kwargs = {}
    if not permissive:
        field_kwargs = {}
        if min_value is not None:
            field_kwargs["ge"] = min_value
        if max_value is not None:
            field_kwargs["le"] = max_value

    class ConstrainedNumericResponse(BaseModel):
        answer: Union[int, float] = Field(**field_kwargs)
        comment: Optional[str] = Field(None)
        generated_tokens: Optional[Any] = Field(None)

    return ConstrainedNumericResponse


class NumericalResponseValidator(ResponseValidatorABC):
    required_params = ["min_value", "max_value", "permissive"]

    valid_examples = [
        ({"answer": 1}, {"min_value": 0, "max_value": 10}),
        ({"answer": 1}, {"min_value": None, "max_value": None}),
    ]

    invalid_examples = [
        ({"answer": 10}, {"min_value": 0, "max_value": 5}, "Answer is out of range"),
        ({"answer": "ten"}, {"min_value": 0, "max_value": 5}, "Answer is not a number"),
        ({}, {"min_value": 0, "max_value": 5}, "Answer key is missing"),
    ]

    def fix(self, response, verbose=False):
        response_text = str(response).lower()
        import re

        if verbose:
            print(f"Ivalid generated tokens was was: {response_text}")
        pattern = r"\b\d+(?:\.\d+)?\b"
        match = re.search(pattern, response_text.replace(",", ""))
        solution = match.group(0) if match else response.get("answer")
        if verbose:
            print("Proposed solution is: ", solution)
        if "comment" in response:
            return {"answer": solution, "comment": response["comment"]}
        else:
            return {"answer": solution}

    def _check_constraints(self, pydantic_edsl_answer: BaseModel):
        pass


class QuestionNumerical(QuestionBase):
    """This question prompts the agent to answer with a numerical value.

    >>> QuestionNumerical.self_check()

    """

    question_type = "numerical"
    min_value: Optional[float] = NumericalOrNoneDescriptor()
    max_value: Optional[float] = NumericalOrNoneDescriptor()

    _response_model = None
    response_validator_class = NumericalResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        include_comment: bool = True,
        question_presentation: Optional[str] = None,
        answering_instructions: Optional[str] = None,
        permissive: bool = False,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param min_value: The minimum value of the answer.
        :param max_value: The maximum value of the answer.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.min_value = min_value
        self.max_value = max_value

        self.include_comment = include_comment
        self.question_presentation = question_presentation
        self.answering_instructions = answering_instructions
        self.permissive = permissive

    def create_response_model(self):
        return create_numeric_response(self.min_value, self.max_value, self.permissive)

    ################
    # Answer methods
    ################

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <div>
        <input type="number" id="{{ question_name }}" name="{{ question_name }}">
        </div>
        """
        ).render(question_name=self.question_name)
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    @inject_exception
    def example(cls, include_comment=False) -> QuestionNumerical:
        """Return an example question."""
        return cls(
            question_name="age",
            question_text="You are a 45 year old man. How old are you in years?",
            min_value=0,
            max_value=86.7,
            include_comment=include_comment,
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
import random
import textwrap
from typing import Any, Optional, Union
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import IntegerOrNoneDescriptor
from edsl.questions.decorators import inject_exception

from pydantic import field_validator, Field
from edsl.questions.ResponseValidatorABC import ResponseValidatorABC
from edsl.questions.ResponseValidatorABC import BaseResponse

from edsl.exceptions import QuestionAnswerValidationError
import textwrap
import json

from json_repair import repair_json


def convert_string(s):
    """Convert a string to a more appropriate type if possible.

    >>> convert_string("3.14")
    3.14
    >>> convert_string("42")
    42
    >>> convert_string("hello")
    'hello'
    >>> convert_string('{"key": "value"}')
    {'key': 'value'}
    >>> convert_string("{'key': 'value'}")
    {'key': 'value'}
    """

    if not isinstance(s, str):  # if it's not a string, return it as is
        return s

    # If the repair returns, continue on; otherwise, try to load it as JSON
    if (repaired_json := repair_json(s)) == '""':
        pass
    else:
        try:
            return json.loads(repaired_json)
        except json.JSONDecodeError:
            pass

    # Try to convert to float
    try:
        return float(s)
    except ValueError:
        pass

    # Try to convert to int
    try:
        return int(s)
    except ValueError:
        pass

    # If all conversions fail, return the original string
    return s


def create_model(max_list_items: int, permissive):
    from pydantic import BaseModel

    if permissive or max_list_items is None:

        class ListResponse(BaseModel):
            answer: list[Any]
            comment: Optional[str] = None
            generated_tokens: Optional[str] = None

    else:

        class ListResponse(BaseModel):
            """
            >>> nr = ListResponse(answer=["Apple", "Cherry"])
            >>> nr.dict()
            {'answer': ['Apple', 'Cherry'], 'comment': None, 'generated_tokens': None}
            """

            answer: list[Any] = Field(..., min_items=0, max_items=max_list_items)
            comment: Optional[str] = None
            generated_tokens: Optional[str] = None

    return ListResponse


class ListResponseValidator(ResponseValidatorABC):
    required_params = ["max_list_items", "permissive"]
    valid_examples = [({"answer": ["hello", "world"]}, {"max_list_items": 5})]

    invalid_examples = [
        (
            {"answer": ["hello", "world", "this", "is", "a", "test"]},
            {"max_list_items": 5},
            "Too many items.",
        ),
    ]

    def _check_constraints(self, response) -> None:
        if (
            self.max_list_items is not None
            and len(response.answer) > self.max_list_items
        ):
            raise QuestionAnswerValidationError("Too many items.")

    def fix(self, response, verbose=False):
        if verbose:
            print(f"Fixing list response: {response}")
        answer = str(response.get("answer") or response.get("generated_tokens", ""))
        if len(answer.split(",")) > 0:
            return (
                {"answer": answer.split(",")} | {"comment": response.get("comment")}
                if "comment" in response
                else {}
            )

    def _post_process(self, edsl_answer_dict):
        edsl_answer_dict["answer"] = [
            convert_string(item) for item in edsl_answer_dict["answer"]
        ]
        return edsl_answer_dict


class QuestionList(QuestionBase):
    """This question prompts the agent to answer by providing a list of items as comma-separated strings."""

    question_type = "list"
    max_list_items: int = IntegerOrNoneDescriptor()
    _response_model = None
    response_validator_class = ListResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        max_list_items: Optional[int] = None,
        include_comment: bool = True,
        answering_instructions: Optional[str] = None,
        question_presentation: Optional[str] = None,
        permissive: bool = False,
    ):
        """Instantiate a new QuestionList.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param max_list_items: The maximum number of items that can be in the answer list.

        >>> QuestionList.example().self_check()
        """
        self.question_name = question_name
        self.question_text = question_text
        self.max_list_items = max_list_items
        self.permissive = permissive

        self.include_comment = include_comment
        self.answering_instructions = answering_instructions
        self.question_presentations = question_presentation

    def create_response_model(self):
        return create_model(self.max_list_items, self.permissive)

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <div id="question-list-container">
            <div>
                <textarea name="{{ question_name }}[]" rows="1" placeholder="Enter item"></textarea>
            </div>
        </div>
        <button type="button" onclick="addNewLine()">Add another line</button>

        <script>
            function addNewLine() {
                var container = document.getElementById('question-list-container');
                var newLine = document.createElement('div');
                newLine.innerHTML = '<textarea name="{{ question_name }}[]" rows="1" placeholder="Enter item"></textarea>';
                container.appendChild(newLine);
            }
        </script>
        """
        ).render(question_name=self.question_name)
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    @inject_exception
    def example(
        cls, include_comment=True, max_list_items=None, permissive=False
    ) -> QuestionList:
        """Return an example of a list question."""
        return cls(
            question_name="list_of_foods",
            question_text="What are your favorite foods?",
            include_comment=include_comment,
            max_list_items=max_list_items,
            permissive=permissive,
        )


def main():
    """Create an example of a list question and demonstrate its functionality."""
    from edsl.questions.QuestionList import QuestionList

    q = QuestionList.example(max_list_items=5)
    q.question_text
    q.question_name
    q.max_list_items
    # validate an answer
    q._validate_answer({"answer": ["pasta", "garlic", "oil", "parmesan"]})
    # translate answer code
    q._translate_answer_code_to_answer(["pasta", "garlic", "oil", "parmesan"])
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from typing import Optional
from collections import UserList
from edsl import Model

from edsl.language_models import LanguageModel
from edsl.Base import Base
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version
from edsl.utilities.utilities import is_valid_variable_name
from edsl.utilities.utilities import dict_hash


class ModelList(Base, UserList):
    def __init__(self, data: Optional[list] = None):
        """Initialize the ScenarioList class.

        >>> from edsl import Model
        >>> m = ModelList(Model.available())

        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__([])

    @property
    def names(self):
        """

        >>> ModelList.example().names
        {'...'}
        """
        return set([model.model for model in self])

    def rich_print(self):
        pass

    def __repr__(self):
        return f"ModelList({super().__repr__()})"

    def __hash__(self):
        """Return a hash of the ModelList. This is used for comparison of ModelLists.

        >>> isinstance(hash(Model()), int)
        True

        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self._to_dict(sort=True))

    def _to_dict(self, sort=False):
        if sort:
            model_list = sorted([model for model in self], key=lambda x: hash(x))
            return {"models": [model._to_dict() for model in model_list]}
        else:
            return {"models": [model._to_dict() for model in self]}

    @classmethod
    def from_names(self, *args, **kwargs):
        """A a model list from a list of names"""
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]
        return ModelList([Model(model_name, **kwargs) for model_name in args])

    @add_edsl_version
    def to_dict(self):
        """
        Convert the ModelList to a dictionary.
        >>> ModelList.example().to_dict()
        {'models': [...], 'edsl_version': '...', 'edsl_class_name': 'ModelList'}
        """
        return self._to_dict()

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data):
        """
        Create a ModelList from a dictionary.

        >>> newm = ModelList.from_dict(ModelList.example().to_dict())
        >>> assert ModelList.example() == newm
        """
        return cls(data=[LanguageModel.from_dict(model) for model in data["models"]])

    def code(self):
        pass

    @classmethod
    def example(cls, randomize: bool = False) -> "ModelList":
        """
        Returns an example ModelList instance.

        :param randomize: If True, uses Model's randomize method.
        """

        return cls([Model.example(randomize) for _ in range(3)])


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from abc import abstractmethod, ABC
import os
import re
from datetime import datetime, timedelta
from edsl.config import CONFIG


class InferenceServiceABC(ABC):
    """
    Abstract class for inference services.
    Anthropic: https://docs.anthropic.com/en/api/rate-limits
    """

    _coop_config_vars = None

    default_levels = {
        "google": {"tpm": 2_000_000, "rpm": 15},
        "openai": {"tpm": 2_000_000, "rpm": 10_000},
        "anthropic": {"tpm": 2_000_000, "rpm": 500},
    }

    def __init_subclass__(cls):
        """
        Check that the subclass has the required attributes.
        - `key_sequence` attribute determines...
        - `model_exclude_list` attribute determines...
        """
        if not hasattr(cls, "key_sequence"):
            raise NotImplementedError(
                f"Class {cls.__name__} must have a 'key_sequence' attribute."
            )
        if not hasattr(cls, "model_exclude_list"):
            raise NotImplementedError(
                f"Class {cls.__name__} must have a 'model_exclude_list' attribute."
            )

    @classmethod
    def _should_refresh_coop_config_vars(cls):
        """
        Returns True if config vars have been fetched over 24 hours ago, and False otherwise.
        """

        if cls._last_config_fetch is None:
            return True
        return (datetime.now() - cls._last_config_fetch) > timedelta(hours=24)

    @classmethod
    def _get_limt(cls, limit_type: str) -> int:
        key = f"EDSL_SERVICE_{limit_type.upper()}_{cls._inference_service_.upper()}"
        if key in os.environ:
            return int(os.getenv(key))

        if cls._coop_config_vars is None or cls._should_refresh_coop_config_vars():
            try:
                from edsl import Coop

                c = Coop()
                cls._coop_config_vars = c.fetch_rate_limit_config_vars()
                cls._last_config_fetch = datetime.now()
                if key in cls._coop_config_vars:
                    return cls._coop_config_vars[key]
            except Exception:
                cls._coop_config_vars = None
        else:
            if key in cls._coop_config_vars:
                return cls._coop_config_vars[key]

        if cls._inference_service_ in cls.default_levels:
            return int(cls.default_levels[cls._inference_service_][limit_type])

        return int(CONFIG.get(f"EDSL_SERVICE_{limit_type.upper()}_BASELINE"))

    def get_tpm(cls) -> int:
        """
        Returns the TPM for the service. If the service is not defined in the environment variables, it will return the baseline TPM.
        """
        return cls._get_limt(limit_type="tpm")

    def get_rpm(cls):
        """
        Returns the RPM for the service. If the service is not defined in the environment variables, it will return the baseline RPM.
        """
        return cls._get_limt(limit_type="rpm")

    @abstractmethod
    def available() -> list[str]:
        """
        Returns a list of available models for the service.
        """
        pass

    @abstractmethod
    def create_model():
        """
        Returns a LanguageModel object.
        """
        pass

    @staticmethod
    def to_class_name(s):
        """
        Converts a string to a valid class name.

        >>> InferenceServiceABC.to_class_name("hello world")
        'HelloWorld'
        """

        s = re.sub(r"[^a-zA-Z0-9 ]", "", s)
        s = "".join(word.title() for word in s.split())
        if s and s[0].isdigit():
            s = "Class" + s
        return s


if __name__ == "__main__":
    pass
    # deep_infra_service = DeepInfraService("deep_infra", "DEEP_INFRA_API_KEY")
    # deep_infra_service.available()
    # m = deep_infra_service.create_model("microsoft/WizardLM-2-7B")
    # response = m().hello()
    # print(response)

    # anthropic_service = AnthropicService("anthropic", "ANTHROPIC_API_KEY")
    # anthropic_service.available()
    # m = anthropic_service.create_model("claude-3-opus-20240229")
    # response = m().hello()
    # print(response)
    # factory = OpenAIService("openai", "OPENAI_API")
    # factory.available()
    # m = factory.create_model("gpt-3.5-turbo")
    # response = m().hello()

    # from edsl import QuestionFreeText
    # results = QuestionFreeText.example().by(m()).run()

    # collection = InferenceServicesCollection([
    #     OpenAIService,
    #     AnthropicService,
    #     DeepInfraService
    # ])

    # available = collection.available()
    # factory = collection.create_model_factory(*available[0])
    # m = factory()
    # from edsl import QuestionFreeText
    # results = QuestionFreeText.example().by(m).run()
    # print(results)



from typing import NamedTuple, Dict, List, Optional, Any
from dataclasses import dataclass, fields
import reprlib


class ModelInputs(NamedTuple):
    "This is what was send by the agent to the model"
    user_prompt: str
    system_prompt: str
    encoded_image: Optional[str] = None


class EDSLOutput(NamedTuple):
    "This is the edsl dictionary that is returned by the model"
    answer: Any
    generated_tokens: str
    comment: Optional[str] = None


class ModelResponse(NamedTuple):
    "This is the metadata that is returned by the model and includes info about the cache"
    response: dict
    cache_used: bool
    cache_key: str
    cached_response: Optional[Dict[str, Any]] = None
    cost: Optional[float] = None


class AgentResponseDict(NamedTuple):
    edsl_dict: EDSLOutput
    model_inputs: ModelInputs
    model_outputs: ModelResponse


class EDSLResultObjectInput(NamedTuple):
    generated_tokens: str
    question_name: str
    prompts: dict
    cached_response: str
    raw_model_response: str
    cache_used: bool
    cache_key: str
    answer: Any
    comment: str
    validated: bool = False
    exception_occurred: Exception = None
    cost: Optional[float] = None


@dataclass
class ImageInfo:
    file_path: str
    file_name: str
    image_format: str
    file_size: int
    encoded_image: str

    def __repr__(self):
        reprlib_instance = reprlib.Repr()
        reprlib_instance.maxstring = 30  # Limit the string length for the encoded image

        # Get all fields except encoded_image
        field_reprs = [
            f"{f.name}={getattr(self, f.name)!r}"
            for f in fields(self)
            if f.name != "encoded_image"
        ]

        # Add the reprlib-restricted encoded_image field
        field_reprs.append(f"encoded_image={reprlib_instance.repr(self.encoded_image)}")

        # Join everything to create the repr
        return f"{self.__class__.__name__}({', '.join(field_reprs)})"



"""A module to represent a dataset of observations."""

from __future__ import annotations
import random
import json
from collections import UserList
from typing import Any, Union, Optional

import numpy as np

from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.results.DatasetTree import Tree


class Dataset(UserList, ResultsExportMixin):
    """A class to represent a dataset of observations."""

    def __init__(self, data: list[dict[str, Any]] = None):
        """Initialize the dataset with the given data."""
        super().__init__(data)

    def __len__(self) -> int:
        """Return the number of observations in the dataset.

        Need to override the __len__ method to return the number of observations in the dataset because
        otherwise, the UserList class would return the number of dictionaries in the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> len(d)
        4
        """
        _, values = list(self.data[0].items())[0]
        return len(values)

    def keys(self):
        """Return the keys of the first observation in the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.keys()
        ['a.b']
        """
        return [list(o.keys())[0] for o in self]

    def __repr__(self) -> str:
        """Return a string representation of the dataset."""
        return f"Dataset({self.data})"

    def _key_to_value(self, key: str) -> Any:
        """Retrieve the value associated with the given key from the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d._key_to_value('a.b')
        [1, 2, 3, 4]

        >>> d._key_to_value('a')
        Traceback (most recent call last):
        ...
        KeyError: "Key 'a' not found in any of the dictionaries."

        """
        potential_matches = []
        for data_dict in self.data:
            data_key, data_values = list(data_dict.items())[0]
            if key == data_key:
                return data_values
            if key == data_key.split(".")[-1]:
                potential_matches.append((data_key, data_values))

        if len(potential_matches) == 1:
            return potential_matches[0][1]
        elif len(potential_matches) > 1:
            raise KeyError(
                f"Key '{key}' found in more than one location: {[m[0] for m in potential_matches]}"
            )

        raise KeyError(f"Key '{key}' not found in any of the dictionaries.")

    def first(self) -> dict[str, Any]:
        """Get the first value of the first key in the first dictionary.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.first()
        1
        """

        def get_values(d):
            """Get the values of the first key in the dictionary."""
            return list(d.values())[0]

        return get_values(self.data[0])[0]

    def select(self, *keys):
        """Return a new dataset with only the selected keys.

        :param keys: The keys to select.

        >>> d = Dataset([{'a.b':[1,2,3,4]}, {'c.d':[5,6,7,8]}])
        >>> d.select('a.b')
        Dataset([{'a.b': [1, 2, 3, 4]}])

        >>> d.select('a.b', 'c.d')
        Dataset([{'a.b': [1, 2, 3, 4]}, {'c.d': [5, 6, 7, 8]}])
        """
        if isinstance(keys, str):
            keys = [keys]

        new_data = []
        for observation in self.data:
            observation_key = list(observation.keys())[0]
            if observation_key in keys:
                new_data.append(observation)
        return Dataset(new_data)

    def to_json(self):
        """Return a JSON representation of the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.to_json()
        [{'a.b': [1, 2, 3, 4]}]
        """
        return json.loads(
            json.dumps(self.data)
        )  # janky but I want to make sure it's serializable & deserializable

    def _repr_html_(self) -> str:
        """Return an HTML representation of the dataset."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.data)

    def shuffle(self, seed=None) -> Dataset:
        """Return a new dataset with the observations shuffled.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.shuffle(seed=0)
        Dataset([{'a.b': [3, 1, 2, 4]}])
        """
        if seed is not None:
            random.seed(seed)

        indices = None

        for entry in self:
            key, values = list(entry.items())[0]
            if indices is None:
                indices = list(range(len(values)))
                random.shuffle(indices)
            entry[key] = [values[i] for i in indices]

        return self

    def sample(
        self,
        n: int = None,
        frac: float = None,
        with_replacement: bool = True,
        seed: Union[str, int, float] = None,
    ) -> Dataset:
        """Return a new dataset with a sample of the observations.

        :param n: The number of samples to take.
        :param frac: The fraction of samples to take.
        :param with_replacement: Whether to sample with replacement.
        :param seed: The seed for the random number generator.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.sample(n=2, seed=0, with_replacement=True)
        Dataset([{'a.b': [4, 4]}])

        >>> d.sample(n = 10, seed=0, with_replacement=False)
        Traceback (most recent call last):
        ...
        ValueError: Sample size cannot be greater than the number of available elements when sampling without replacement.
        """
        if seed is not None:
            random.seed(seed)

        # Validate the input for sampling parameters
        if n is None and frac is None:
            raise ValueError("Either 'n' or 'frac' must be provided for sampling.")

        if n is not None and frac is not None:
            raise ValueError("Only one of 'n' or 'frac' should be specified.")

        # Get the length of the lists from the first entry
        first_key, first_values = list(self[0].items())[0]
        total_length = len(first_values)

        # Determine the number of samples based on 'n' or 'frac'
        if n is None:
            n = int(total_length * frac)

        if not with_replacement and n > total_length:
            raise ValueError(
                "Sample size cannot be greater than the number of available elements when sampling without replacement."
            )

        # Sample indices based on the method chosen
        if with_replacement:
            indices = [random.randint(0, total_length - 1) for _ in range(n)]
        else:
            indices = random.sample(range(total_length), k=n)

        # Apply the same indices to all entries
        for entry in self:
            key, values = list(entry.items())[0]
            entry[key] = [values[i] for i in indices]

        return self

    def order_by(self, sort_key: str, reverse: bool = False) -> Dataset:
        """Return a new dataset with the observations sorted by the given key.

        :param sort_key: The key to sort the observations by.
        :param reverse: Whether to sort in reverse order.

        >>> d = Dataset([{'a':[1,2,3,4]}, {'b':[4,3,2,1]}])
        >>> d.order_by('a')
        Dataset([{'a': [1, 2, 3, 4]}, {'b': [4, 3, 2, 1]}])

        >>> d.order_by('a', reverse=True)
        Dataset([{'a': [4, 3, 2, 1]}, {'b': [1, 2, 3, 4]}])

        >>> d = Dataset([{'X.a':[1,2,3,4]}, {'X.b':[4,3,2,1]}])
        >>> d.order_by('a')
        Dataset([{'X.a': [1, 2, 3, 4]}, {'X.b': [4, 3, 2, 1]}])


        """

        def sort_indices(lst: list[Any]) -> list[int]:
            """
            Return the indices that would sort the list.

            :param lst: The list to be sorted.
            :return: A list of indices that would sort the list.
            """
            indices = np.argsort(lst).tolist()
            if reverse:
                indices.reverse()
            return indices

        number_found = 0
        for obs in self.data:
            key, values = list(obs.items())[0]
            # an obseration is {'a':[1,2,3,4]}
            # key = list(obs.keys())[0]
            if (
                sort_key == key or sort_key == key.split(".")[-1]
            ):  # e.g., "age" in "scenario.age"
                relevant_values = values
                number_found += 1

        if number_found == 0:
            raise ValueError(f"Key '{sort_key}' not found in any of the dictionaries.")
        elif number_found > 1:
            raise ValueError(f"Key '{sort_key}' found in more than one dictionary.")

        # relevant_values = self._key_to_value(sort_key)
        sort_indices_list = sort_indices(relevant_values)
        new_data = []
        for observation in self.data:
            # print(observation)
            key, values = list(observation.items())[0]
            new_values = [values[i] for i in sort_indices_list]
            new_data.append({key: new_values})

        return Dataset(new_data)

    @property
    def tree(self):
        """Return a tree representation of the dataset.

        >>> d = Dataset([{'a':[1,2,3,4]}, {'b':[4,3,2,1]}])
        >>> d.tree.print_tree()
        Tree has not been constructed yet.
        """
        return Tree(self)

    @classmethod
    def example(self):
        """Return an example dataset.

        >>> Dataset.example()
        Dataset([{'a': [1, 2, 3, 4]}, {'b': [4, 3, 2, 1]}])
        """
        return Dataset([{"a": [1, 2, 3, 4]}, {"b": [4, 3, 2, 1]}])


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



class RemoteCacheSync:
    def __init__(
        self, coop, cache, output_func, remote_cache=True, remote_cache_description=""
    ):
        self.coop = coop
        self.cache = cache
        self._output = output_func
        self.remote_cache = remote_cache
        self.old_entry_keys = []
        self.new_cache_entries = []
        self.remote_cache_description = remote_cache_description

    def __enter__(self):
        if self.remote_cache:
            self._sync_from_remote()
            self.old_entry_keys = list(self.cache.keys())
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.remote_cache:
            self._sync_to_remote()
        return False  # Propagate exceptions

    def _sync_from_remote(self):
        cache_difference = self.coop.remote_cache_get_diff(self.cache.keys())
        client_missing_cacheentries = cache_difference.get(
            "client_missing_cacheentries", []
        )
        missing_entry_count = len(client_missing_cacheentries)

        if missing_entry_count > 0:
            self._output(
                f"Updating local cache with {missing_entry_count:,} new "
                f"{'entry' if missing_entry_count == 1 else 'entries'} from remote..."
            )
            self.cache.add_from_dict(
                {entry.key: entry for entry in client_missing_cacheentries}
            )
            self._output("Local cache updated!")
        else:
            self._output("No new entries to add to local cache.")

    def _sync_to_remote(self):
        cache_difference = self.coop.remote_cache_get_diff(self.cache.keys())
        server_missing_cacheentry_keys = cache_difference.get(
            "server_missing_cacheentry_keys", []
        )
        server_missing_cacheentries = [
            entry
            for key in server_missing_cacheentry_keys
            if (entry := self.cache.data.get(key)) is not None
        ]

        new_cache_entries = [
            entry
            for entry in self.cache.values()
            if entry.key not in self.old_entry_keys
        ]
        server_missing_cacheentries.extend(new_cache_entries)
        new_entry_count = len(server_missing_cacheentries)

        if new_entry_count > 0:
            self._output(
                f"Updating remote cache with {new_entry_count:,} new "
                f"{'entry' if new_entry_count == 1 else 'entries'}..."
            )
            self.coop.remote_cache_create_many(
                server_missing_cacheentries,
                visibility="private",
                description=self.remote_cache_description,
            )
            self._output("Remote cache updated!")
        else:
            self._output("No new entries to add to remote cache.")

        self._output(
            f"There are {len(self.cache.keys()):,} entries in the local cache."
        )


# # Usage example
# def run_job(self, n, progress_bar, cache, stop_on_exception, sidecar_model, print_exceptions, raise_validation_errors, use_remote_cache=True):
#     with RemoteCacheSync(self.coop, cache, self._output, remote_cache=use_remote_cache):
#         self._output("Running job...")
#         results = self._run_local(
#             n=n,
#             progress_bar=progress_bar,
#             cache=cache,
#             stop_on_exception=stop_on_exception,
#             sidecar_model=sidecar_model,
#             print_exceptions=print_exceptions,
#             raise_validation_errors=raise_validation_errors,
#         )
#         self._output("Job completed!")

#     results.cache = cache.new_entries_cache()
#     return results



from __future__ import annotations
import ast
import json
import os
import shutil
import sqlite3
from edsl.config import CONFIG
from edsl.data.Cache import Cache
from edsl.data.CacheEntry import CacheEntry
from edsl.data.SQLiteDict import SQLiteDict

from edsl.config import CONFIG


def set_session_cache(cache: Cache) -> None:
    """
    Set the session cache.
    """
    CONFIG.EDSL_SESSION_CACHE = cache


def unset_session_cache() -> None:
    """
    Unset the session cache.
    """
    if hasattr(CONFIG, "EDSL_SESSION_CACHE"):
        del CONFIG.EDSL_SESSION_CACHE


class CacheHandler:
    """
    This CacheHandler figures out what caches are available and does migrations, as needed.
    """

    CACHE_PATH = CONFIG.get("EDSL_DATABASE_PATH")

    def __init__(self, test: bool = False):
        self.test = test
        self.create_cache_directory()
        self.cache = self.gen_cache()
        old_data = self.from_old_sqlite_cache()
        self.cache.add_from_dict(old_data)

    def create_cache_directory(self, notify=False) -> None:
        """
        Create the cache directory if one is required and it does not exist.
        """
        path = self.CACHE_PATH.replace("sqlite:///", "")
        dir_path = os.path.dirname(path)
        if dir_path and not os.path.exists(dir_path):
            os.makedirs(dir_path)
            if notify:
                print(f"Created cache directory: {dir_path}")

    def gen_cache(self) -> Cache:
        """
        Generate a Cache object.
        """
        if self.test:
            return Cache(data={})

        if hasattr(CONFIG, "EDSL_SESSION_CACHE"):
            return CONFIG.EDSL_SESSION_CACHE

        cache = Cache(data=SQLiteDict(self.CACHE_PATH))
        return cache

    def from_old_sqlite_cache(
        self, path: str = "edsl_cache.db"
    ) -> dict[str, CacheEntry]:
        """
        Convert an old-style cache to the new format.
        - NB: Not worth converting to sqlalchemy - this is a one-time operation.
        """
        old_data = {}
        if not os.path.exists(os.path.join(os.getcwd(), path)):
            return old_data
        try:
            conn = sqlite3.connect(path)
            with conn:
                cur = conn.cursor()
                table_name = "responses"
                cur.execute(f"PRAGMA table_info({table_name})")
                columns = cur.fetchall()
                schema = {column[1]: column[2] for column in columns}
                data = cur.execute(f"SELECT * FROM {table_name}").fetchall()
                for row in data:
                    entry = self._parse_old_cache_entry(row, schema)
                    old_data[entry.key] = entry
            print(
                f"Found old cache at {path} with {len(old_data)} entries.\n"
                f"We will convert this to the new cache format.\n"
                f"The old cache is backed up to {path}.bak"
            )
            shutil.copy(path, f"{path}.bak")
            os.remove(path)
        except sqlite3.OperationalError:
            print("Found an old Cache but could not convert it to new format.")

        return old_data

    def _parse_old_cache_entry(self, row: tuple, schema) -> CacheEntry:
        """
        Parse an old cache entry.
        """
        entry_dict = {k: row[i] for i, k in enumerate(schema.keys())}
        _ = entry_dict.pop("id")
        entry_dict["user_prompt"] = entry_dict.pop("prompt")
        parameters = entry_dict["parameters"]
        entry_dict["parameters"] = ast.literal_eval(parameters)
        entry = CacheEntry(**entry_dict)
        return entry

    def get_cache(self) -> Cache:
        return self.cache

    ###############
    # NOT IN USE
    ###############
    def from_sqlite(uri="new_edsl_cache.db") -> dict[str, CacheEntry]:
        """
        Read in a new-style sqlite cache and return a dictionary of dictionaries.
        """
        conn = sqlite3.connect(uri)
        with conn:
            cur = conn.cursor()
            data = cur.execute("SELECT key, value FROM data").fetchall()
            newdata = {}
            for _, value in data:
                entry = CacheEntry.from_dict(json.loads(value))
                newdata[entry.key] = entry
        return newdata

    def from_jsonl(filename="edsl_cache.jsonl") -> dict[str, CacheEntry]:
        """Read in a jsonl file and return a dictionary of CacheEntry objects."""
        with open(filename, "a+") as f:
            f.seek(0)
            lines = f.readlines()
        newdata = {}
        for line in lines:
            d = json.loads(line)
            key = list(d.keys())[0]
            value = list(d.values())[0]
            newdata[key] = CacheEntry.from_dict(value)
        return newdata


if __name__ == "__main__":
    ch = CacheHandler()



from edsl.conjure.InputDataPyRead import InputDataPyRead


class InputDataSPSS(InputDataPyRead):
    def pyread_function(self, datafile_name):
        from pyreadstat import read_sav

        return read_sav(datafile_name)



from textwrap import dedent
from dataclasses import dataclass

from typing import List

from edsl.auto.StageBase import StageBase
from edsl.auto.StageBase import FlowDataBase

from edsl.auto.StagePersona import StagePersona

from edsl.questions import QuestionList
from edsl.scenarios import Scenario
from edsl import Model

from edsl.auto.utilities import gen_pipeline


class StagePersonaDimensions(StageBase):
    input = StagePersona.output

    @dataclass
    class Output(FlowDataBase):
        attribute_results: List[str]
        persona: str

    output = Output

    def handle_data(self, data):
        q_attributes = QuestionList(
            question_text=dedent(
                """\
            Here is a persona: "{{ persona }}"
            It was construced to be someone who could answer these questions: "{{ questions }}"

            We want to identify the general dimensions that make up this persona.
            E.g., if the person is desribed as 'happy' then a dimenion would be  'mood'
            """
            ),
            question_name="find_attributes",
        )
        m = Model()
        results = (
            q_attributes.by(
                Scenario({"persona": data.persona, "questions": data.questions})
            )
            .by(m)
            .run()
        )
        (
            results.select("find_attributes").print(
                pretty_labels={
                    "answer.find_attributes": f'Persona dimensions for: "{data.persona}"'
                },
                split_at_dot=False,
            )
        )
        attribute_results = results.select("find_attributes").first()
        return self.output(attribute_results=attribute_results, persona=data.persona)


if __name__ == "__main__":
    from edsl.auto.StageQuestions import StageQuestions

    pipeline = gen_pipeline([StageQuestions, StagePersona, StagePersonaDimensions])
    pipeline.process(
        pipeline.input(
            overall_question="What are some factors that could determine whether someone likes ice cream?"
        )
    )



"""Mixin class for exporting results."""

from functools import wraps
from edsl.results.DatasetExportMixin import DatasetExportMixin


def to_dataset(func):
    """Convert the object to a Dataset object before calling the function."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Return the function with the Results object converted to a Dataset object."""
        if self.__class__.__name__ == "ScenarioList":
            return func(self.to_dataset(), *args, **kwargs)
        else:
            raise Exception(
                f"Class {self.__class__.__name__} not recognized as a Results or Dataset object."
            )

    return wrapper


def decorate_methods_from_mixin(cls, mixin_cls):
    for attr_name, attr_value in mixin_cls.__dict__.items():
        if callable(attr_value) and not attr_name.startswith("__"):
            setattr(cls, attr_name, to_dataset(attr_value))
    return cls


# def decorate_all_methods(cls):
#     for attr_name, attr_value in cls.__dict__.items():
#         if callable(attr_value):
#             setattr(cls, attr_name, to_dataset(attr_value))
#     return cls


# @decorate_all_methods
class ScenarioListExportMixin(DatasetExportMixin):
    """Mixin class for exporting Results objects."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        decorate_methods_from_mixin(cls, DatasetExportMixin)

    def to_docx(self, filename: str):
        """Export the ScenarioList to a .docx file."""
        dataset = self.to_dataset()
        from edsl.results.DatasetTree import Tree

        tree = Tree(dataset)
        tree.construct_tree()
        tree.to_docx(filename)



from typing import Dict, List, Any, Optional
from docx import Document
from docx.shared import Inches, Pt
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.enum.style import WD_STYLE_TYPE


class TreeNode:
    def __init__(self, key=None, value=None):
        self.key = key
        self.value = value
        self.children = {}


class Tree:
    def __init__(self, data: "Dataset"):
        d = {}
        for entry in data:
            d.update(entry)
        self.data = d
        self.root = None

    def unique_values_by_keys(self) -> dict:
        unique_values = {}
        for key, values in self.data.items():
            unique_values[key] = list(set(values))
        return unique_values

    def construct_tree(self, node_order: Optional[List[str]] = None):
        # Validate node_order
        if node_order is None:
            unique_values = self.unique_values_by_keys()
            # Sort keys by number of unique values
            node_order = sorted(
                unique_values, key=lambda k: len(unique_values[k]), reverse=True
            )
        else:
            if not set(node_order).issubset(set(self.data.keys())):
                invalid_keys = set(node_order) - set(self.data.keys())
                raise ValueError(f"Invalid keys in node_order: {invalid_keys}")

        self.root = TreeNode()

        for i in range(len(self.data[node_order[0]])):
            current = self.root
            for level in node_order[:-1]:
                value = self.data[level][i]
                if value not in current.children:
                    current.children[value] = TreeNode(key=level, value=value)
                current = current.children[value]

            leaf_key = node_order[-1]
            leaf_value = self.data[leaf_key][i]
            if leaf_value not in current.children:
                current.children[leaf_value] = TreeNode(key=leaf_key, value=leaf_value)

    def print_tree(
        self, node: Optional[TreeNode] = None, level: int = 0, print_keys: bool = False
    ):
        if node is None:
            node = self.root
            if node is None:
                print("Tree has not been constructed yet.")
                return

        if node.value is not None:
            if print_keys and node.key is not None:
                print("  " * level + f"{node.key}: {node.value}")
            else:
                print("  " * level + str(node.value))
        for child in node.children.values():
            self.print_tree(child, level + 1, print_keys)

    def to_docx(self, filename: str):
        doc = Document()

        # Create styles for headings
        for i in range(1, 10):  # Up to 9 levels of headings
            style_name = f"Heading {i}"
            if style_name not in doc.styles:
                doc.styles.add_style(style_name, WD_STYLE_TYPE.PARAGRAPH)

        # Get or create the 'Body Text' style
        if "Body Text" not in doc.styles:
            body_style = doc.styles.add_style("Body Text", WD_STYLE_TYPE.PARAGRAPH)
        else:
            body_style = doc.styles["Body Text"]

        body_style.font.size = Pt(11)

        self._add_to_docx(doc, self.root, 0)
        doc.save(filename)

    def _add_to_docx(self, doc, node: TreeNode, level: int):
        if node.value is not None:
            if level == 0:
                doc.add_heading(str(node.value), level=level + 1)
            elif node.children:  # If the node has children, it's not the last level
                para = doc.add_paragraph(str(node.value))
                para.style = f"Heading {level+1}"
            else:  # If the node has no children, it's the last level (body text)
                para = doc.add_paragraph(str(node.value))
                para.style = "Body Text"

        # Process child nodes (moved outside the if block)
        for child in node.children.values():
            self._add_to_docx(doc, child, level + 1)


# Example usage (commented out)
"""
from edsl.results.Dataset import Dataset

data = Dataset(
    [
        {"continent": ["North America", "Asia", "Europe", "North America", "Asia"]},
        {"country": ["US", "China", "France", "Canada", "Japan"]},
        {"city": ["New York", "Beijing", "Paris", "Toronto", "Tokyo"]},
        {"population": [8419000, 21540000, 2161000, 2930000, 13960000]},
    ]
)

tree = Tree(data)

try:
    tree.construct_tree(["continent", "country", "city", "population"])
    print("Tree without key names:")
    tree.print_tree()
    print("\nTree with key names:")
    tree.print_tree(print_keys=True)
except ValueError as e:
    print(f"Error: {e}")

# Demonstrating validation
try:
    tree.construct_tree(["continent", "country", "invalid_key"])
except ValueError as e:
    print(f"\nValidation Error: {e}")

tree = Tree(data)
tree.construct_tree(["continent", "country", "city", "population"])
tree.print_tree(print_keys=True)
tree.to_docx("tree_structure.docx")
print("DocX file 'tree_structure.docx' has been created.")
"""
